"""This module defines the TeambuilderPokemon class, which is used as an intermediate
format to specify pokemon builds in teambuilders custom classes.
"""

from typing import List, Optional

from poke_env.data import to_id_str
from poke_env.stats import STATS_TO_IDX


class TeambuilderPokemon:
    HP_TO_IVS = {
        "bug": [31, 31, 31, 30, 31, 30],
        "dark": [31, 31, 31, 31, 31, 31],
        "dragon": [30, 31, 31, 31, 31, 31],
        "electric": [31, 31, 31, 31, 30, 31],
        "fighting": [31, 31, 30, 30, 30, 30],
        "fire": [31, 30, 31, 30, 31, 30],
        "flying": [31, 31, 31, 30, 30, 30],
        "ghost": [31, 30, 31, 31, 31, 30],
        "grass": [30, 31, 31, 31, 30, 31],
        "ground": [31, 31, 31, 31, 30, 30],
        "ice": [31, 30, 30, 31, 31, 31],
        "poison": [31, 31, 30, 31, 30, 30],
        "psychic": [30, 31, 31, 30, 31, 31],
        "rock": [31, 31, 30, 30, 31, 30],
        "steel": [31, 31, 31, 31, 31, 30],
        "water": [31, 31, 31, 30, 30, 31],
    }
    evs: List[int]
    ivs: List[int]
    moves: List[str]

    def __init__(
        self,
        nickname: Optional[str] = None,
        species: Optional[str] = None,
        item: Optional[str] = None,
        ability: Optional[str] = None,
        moves: Optional[List[str]] = None,
        nature: Optional[str] = None,
        evs: Optional[List[int]] = None,
        gender: Optional[str] = None,
        ivs: Optional[List[int]] = None,
        shiny: Optional[bool] = None,
        level: Optional[int] = None,
        happiness: Optional[int] = None,
        hiddenpowertype: Optional[str] = None,
        gmax: Optional[bool] = None,
        tera_type: Optional[str] = None,
    ):
        self.nickname = nickname
        self.species = species
        self.item = item
        self.ability = ability
        self.nature = nature
        self.gender = gender
        self.shiny = shiny
        self.level = level
        self.happiness = happiness
        self.hiddenpowertype = hiddenpowertype
        self.gmax = gmax
        self.tera_type = tera_type
        self.evs = evs if evs is not None else [0] * 6
        self.ivs = ivs if ivs is not None else [31] * 6

        if moves is None:
            self.moves = []
        else:
            self.moves = moves

    def __repr__(self) -> str:
        return self.packed

    def __str__(self) -> str:
        return self.packed

    @property
    def packed_evs(self) -> str:
        f_evs = ",".join([str(el) if el != 0 else "" for el in self.evs])
        if f_evs == "," * 5:
            return ""
        return f_evs

    @property
    def packed_ivs(self) -> str:
        f_ivs = ",".join([str(iv) if iv != 31 else "" for iv in self.ivs])
        if f_ivs == "," * 5:
            return ""
        return f_ivs

    @property
    def packed_moves(self) -> str:
        return ",".join([to_id_str(move) for move in self.moves])

    @property
    def packed_endstring(self) -> str:
        f_str = f",{self.hiddenpowertype or ''},"

        if self.gmax:
            return f_str + ",G"
        elif self.tera_type:
            return f_str + f",,,{self.tera_type}"

        if self.hiddenpowertype:
            return f_str

        return ""

    @property
    def packed(self) -> str:
        self._prepare_for_formatting()
        return "%s|%s|%s|%s|%s|%s|%s|%s|%s|%s|%s|%s%s" % (
            self.nickname or "",
            to_id_str(self.species) if self.species else "",
            to_id_str(self.item) if self.item else "",
            to_id_str(self.ability) if self.ability else "",
            self.packed_moves or "",
            self.nature or "",
            self.packed_evs or "",
            self.gender or "",
            self.packed_ivs or "",
            "S" if self.shiny else "",
            self.level or "",
            self.happiness or "",
            self.packed_endstring,
        )

    def _prepare_for_formatting(self):
        for move in self.moves:
            move = to_id_str(move)
            if (
                move.startswith("hiddenpower")
                and len(move) > 11
                and all([iv == 31 for iv in self.ivs])
            ):
                self.ivs = list(self.HP_TO_IVS[move[11:]])

    @staticmethod
    def from_packed(packed_mon: str) -> "TeambuilderPokemon":
        """Converts a packed-format pokemon string into a TeambuilderPokemon object.

        :param packed_mon: The packed-format pokemon string to convert.
        :type packed_mon: str
        :return: The converted TeambuilderPokemon object.
        :rtype: TeambuilderPokemon
        """
        (
            raw_nickname,
            raw_species,
            raw_item,
            raw_ability,
            raw_moves,
            raw_nature,
            raw_evs,
            raw_gender,
            raw_ivs,
            raw_shiny,
            raw_level,
            endstring,
        ) = packed_mon.split("|")

        gmax = False
        tera_type = None
        hiddenpowertype = None
        happiness = None

        if endstring:
            split_endstring = endstring.split(",")

            if split_endstring[0]:
                happiness = int(split_endstring[0])

            if len(split_endstring) == 1:
                pass
            elif split_endstring[-1] == "G":
                gmax = True
            elif split_endstring[-1] != "":
                tera_type = split_endstring[-1]
            elif len(split_endstring) >= 3:
                hiddenpowertype = split_endstring[-2]

        nickname = raw_nickname or None
        species = raw_species or None
        item = raw_item or None
        ability = raw_ability or None
        nature = raw_nature or None
        gender = raw_gender or None

        if raw_moves:
            moves = raw_moves.split(",")
        else:
            moves = None

        if raw_evs:
            evs = [int(ev) if ev else 0 for ev in raw_evs.split(",")]
        else:
            evs = None

        if raw_ivs:
            ivs = [int(iv) if iv else 31 for iv in raw_ivs.split(",")]
        else:
            ivs = None

        if raw_shiny:
            assert raw_shiny == "S"
            shiny = True
        else:
            shiny = False

        if raw_level:
            level = int(raw_level)
        else:
            level = None

        return TeambuilderPokemon(
            nickname=nickname,
            species=species,
            item=item,
            ability=ability,
            moves=moves,
            nature=nature,
            evs=evs,
            gender=gender,
            ivs=ivs,
            shiny=shiny,
            level=level,
            happiness=happiness,
            hiddenpowertype=hiddenpowertype,
            gmax=gmax,
            tera_type=tera_type,
        )

    @staticmethod
    def from_showdown(showdown_mon: str) -> "TeambuilderPokemon":
        """Converts a showdown-format pokemon string into a TeambuilderPokemon object.

        :param showdown_mon: The showdown-format pokemon string to convert.
        :type showdown_mon: str
        :return: The converted TeambuilderPokemon object.
        :rtype: TeambuilderPokemon
        """
        mon = TeambuilderPokemon()

        for line in showdown_mon.split("\n"):
            while line and line.startswith(" "):
                line = line[1:]

            if not line:
                continue
            elif line.startswith("Ability"):
                ability = line.replace("Ability: ", "")
                mon.ability = ability.strip()
            elif line.startswith("Level: "):
                level = line.replace("Level: ", "")
                mon.level = int(level.strip())
            elif line.startswith("Happiness: "):
                happiness = line.replace("Happiness: ", "")
                mon.happiness = int(happiness.strip())
            elif line.startswith("EVs: "):
                evs = line.replace("EVs: ", "")
                split_evs = evs.split(" / ")
                for ev in split_evs:
                    n, stat = ev.split(" ")[:2]
                    idx = STATS_TO_IDX[stat.lower()]
                    mon.evs[idx] = int(n)
            elif line.startswith("IVs: "):
                ivs = line.replace("IVs: ", "")
                ivs_split = ivs.split(" / ")
                for iv in ivs_split:
                    n, stat = iv.split(" ")[:2]
                    idx = STATS_TO_IDX[stat.lower()]
                    mon.ivs[idx] = int(n)
            elif line.startswith("- "):
                line = line.replace("- ", "").strip()
                mon.moves.append(line)
            elif line.startswith("Shiny"):
                mon.shiny = line.strip().endswith("Yes")
            elif line.startswith("Gigantamax"):
                mon.gmax = line.strip().endswith("Yes")
            elif line.strip().endswith(" Nature"):
                nature = line.strip().replace(" Nature", "")
                mon.nature = nature
            elif line.startswith("Hidden Power: "):
                hp_type = line.replace("Hidden Power: ", "").strip()
                mon.hiddenpowertype = hp_type
            elif line.startswith("Tera Type: "):
                tera_type = line.replace("Tera Type: ", "").strip()
                mon.tera_type = tera_type
            else:
                if "@" in line:
                    mon_info, item = line.split(" @ ")
                    mon.item = item.strip()
                else:
                    mon_info = line
                split_mon_info = mon_info.split(" ")
                if split_mon_info[-1] == "(M)":
                    mon.gender = "M"
                    split_mon_info.pop()
                if split_mon_info[-1] == "(F)":
                    mon.gender = "F"
                    split_mon_info.pop()
                if split_mon_info[-1].endswith(")"):
                    for i, info in enumerate(split_mon_info):
                        if info[0] == "(":
                            mon.species = " ".join(split_mon_info[i:])[1:-1]
                            split_mon_info = split_mon_info[:i]
                            break
                mon.nickname = " ".join(split_mon_info)

        return mon



"""This module defines a player class with the Gymnasium API on the main thread.
For a black-box implementation consider using the module env_player.
"""

import asyncio
import time
from abc import abstractmethod
from concurrent.futures import Future
from threading import Thread
from typing import Any, Awaitable, Dict, Generic, List, Optional, Tuple, TypeVar, Union
from weakref import WeakKeyDictionary

import numpy as np
from gymnasium import spaces
from gymnasium.spaces import Space
from gymnasium.spaces.utils import flatdim
from gymnasium.utils import seeding
from numpy.random import Generator
from pettingzoo.utils.env import ParallelEnv  # type: ignore[import-untyped]

from poke_env.battle.abstract_battle import AbstractBattle
from poke_env.battle.battle import Battle
from poke_env.battle.double_battle import DoubleBattle
from poke_env.battle.pokemon import Pokemon
from poke_env.concurrency import create_in_poke_loop
from poke_env.player.battle_order import (
    BattleOrder,
    DoubleBattleOrder,
    ForfeitBattleOrder,
    _EmptyBattleOrder,
)
from poke_env.player.player import Player
from poke_env.ps_client import AccountConfiguration
from poke_env.ps_client.server_configuration import (
    LocalhostServerConfiguration,
    ServerConfiguration,
)
from poke_env.teambuilder.teambuilder import Teambuilder

ItemType = TypeVar("ItemType")
ActionType = TypeVar("ActionType")


class _AsyncQueue(Generic[ItemType]):
    queue: asyncio.Queue[ItemType]

    def __init__(self, loop: asyncio.AbstractEventLoop, maxsize: int = 0):
        self.queue = create_in_poke_loop(asyncio.Queue, loop, maxsize)
        self._loop = loop

    async def async_get(self) -> ItemType:
        return await self.queue.get()

    def get(self, timeout: Optional[float] = None) -> ItemType:
        res = asyncio.run_coroutine_threadsafe(
            asyncio.wait_for(self.async_get(), timeout), self._loop
        )
        return res.result()

    def race_get(self, *events: asyncio.Event) -> Optional[ItemType]:
        async def _race_get() -> Optional[ItemType]:
            get_task = asyncio.create_task(self.async_get())
            wait_tasks = [asyncio.create_task(e.wait()) for e in events]
            done, pending = await asyncio.wait(
                {get_task, *wait_tasks}, return_when=asyncio.FIRST_COMPLETED
            )
            for p in pending:
                p.cancel()
            if get_task in done:
                return get_task.result()
            else:
                return None

        res = asyncio.run_coroutine_threadsafe(_race_get(), self._loop)
        return res.result()

    async def async_put(self, item: ItemType):
        await self.queue.put(item)

    def put(self, item: ItemType):
        self._loop.call_soon_threadsafe(self.queue.put_nowait, item)

    def empty(self):
        return self.queue.empty()


class _EnvPlayer(Player):
    battle_queue: _AsyncQueue[AbstractBattle]
    order_queue: _AsyncQueue[BattleOrder]

    def __init__(
        self, *args: Any, choose_on_teampreview: bool | None = None, **kwargs: Any
    ):
        super().__init__(*args, **kwargs)
        if choose_on_teampreview is None:
            self.logger.warning(
                "choose_on_teampreview arg was not set in environment - by default, teampreview decisions will be made randomly."
            )
        self._choose_on_teampreview = choose_on_teampreview or False
        self.battle_queue = _AsyncQueue(self.ps_client.loop, maxsize=1)
        self.order_queue = _AsyncQueue(self.ps_client.loop, maxsize=1)
        self.battle: Optional[AbstractBattle] = None

    def choose_move(self, battle: AbstractBattle) -> Awaitable[BattleOrder]:
        return self._choose_move(battle)

    async def _choose_move(self, battle: AbstractBattle) -> BattleOrder:
        if not self.battle or self.battle.finished:
            self.battle = battle
        assert self.battle.battle_tag == battle.battle_tag
        await self.battle_queue.async_put(battle)
        order = await self.order_queue.async_get()
        return order

    def teampreview(self, battle: AbstractBattle) -> Awaitable[str]:
        return self._teampreview(battle)

    async def _teampreview(self, battle: AbstractBattle) -> str:
        if not self._choose_on_teampreview:
            return self.random_teampreview(battle)
        elif isinstance(battle, Battle):
            return self.random_teampreview(battle)
        elif isinstance(battle, DoubleBattle):
            if battle.format is None or "vgc" not in battle.format:
                return self.random_teampreview(battle)
            species = [p.base_species for p in battle.team.values()]
            # derive first pair of teampreview selections from first order
            order1 = await self._choose_move(battle)
            if isinstance(order1, (ForfeitBattleOrder, _EmptyBattleOrder)):
                return order1.message
            assert isinstance(order1, DoubleBattleOrder)
            assert isinstance(order1.first_order.order, Pokemon)
            assert isinstance(order1.second_order.order, Pokemon)
            action1 = species.index(order1.first_order.order.base_species) + 1
            action2 = species.index(order1.second_order.order.base_species) + 1
            list(battle.team.values())[action1 - 1]._selected_in_teampreview = True
            list(battle.team.values())[action2 - 1]._selected_in_teampreview = True
            # derive second pair of teampreview selections from second order
            order2 = await self._choose_move(battle)
            if isinstance(order2, (ForfeitBattleOrder, _EmptyBattleOrder)):
                return order2.message
            assert isinstance(order2, DoubleBattleOrder)
            assert isinstance(order2.first_order.order, Pokemon)
            assert isinstance(order2.second_order.order, Pokemon)
            action3 = species.index(order2.first_order.order.base_species) + 1
            action4 = species.index(order2.second_order.order.base_species) + 1
            list(battle.team.values())[action3 - 1]._selected_in_teampreview = True
            list(battle.team.values())[action4 - 1]._selected_in_teampreview = True
            return f"/team {action1}{action2}{action3}{action4}"
        else:
            raise TypeError()

    def _battle_finished_callback(self, battle: AbstractBattle):
        self.battle_queue.queue.put_nowait(battle)


class PokeEnv(ParallelEnv[str, Dict[str, Any], ActionType]):
    """
    Base class implementing the PettingZoo API on the main thread.
    """

    def __init__(
        self,
        *,
        account_configuration1: Optional[AccountConfiguration] = None,
        account_configuration2: Optional[AccountConfiguration] = None,
        avatar: Optional[int] = None,
        battle_format: str = "gen8randombattle",
        log_level: Optional[int] = None,
        save_replays: Union[bool, str] = False,
        server_configuration: Optional[
            ServerConfiguration
        ] = LocalhostServerConfiguration,
        accept_open_team_sheet: Optional[bool] = False,
        start_timer_on_battle_start: bool = False,
        start_listening: bool = True,
        open_timeout: Optional[float] = 10.0,
        ping_interval: Optional[float] = 20.0,
        ping_timeout: Optional[float] = 20.0,
        challenge_timeout: Optional[float] = 60.0,
        team: Optional[Union[str, Teambuilder]] = None,
        choose_on_teampreview: bool | None = None,
        fake: bool = False,
        strict: bool = True,
    ):
        """
        :param account_configuration: Player configuration. If empty, defaults to an
            automatically generated username with no password. This option must be set
            if the server configuration requires authentication.
        :type account_configuration: AccountConfiguration, optional
        :param avatar: Player avatar id. Optional.
        :type avatar: int, optional
        :param battle_format: Name of the battle format this player plays. Defaults to
            gen8randombattle.
        :type battle_format: Optional, str. Default to randombattles, with specifics
            varying per class.
        :param log_level: The player's logger level.
        :type log_level: int. Defaults to logging's default level.
        :param save_replays: Whether to save battle replays. Can be a boolean, where
            True will lead to replays being saved in a potentially new /replay folder,
            or a string representing a folder where replays will be saved.
        :type save_replays: bool or str
        :param server_configuration: Server configuration. Defaults to Localhost Server
            Configuration.
        :type server_configuration: ServerConfiguration, optional
        :param start_listening: Whether to start listening to the server. Defaults to
            True.
        :type start_listening: bool
        :param accept_open_team_sheet: Whether to automatically start the battle with
            open team sheets on. Defaults to False.
        :param start_timer_on_battle_start: Whether to automatically start the battle
            timer on battle start. Defaults to False.
        :type start_timer_on_battle_start: bool
        :param open_timeout: How long to wait for a timeout when connecting the socket
            (important for backend websockets.
            Increase only if timeouts occur during runtime).
            If None connect will never time out.
        :type open_timeout: float, optional
        :param ping_interval: How long between keepalive pings (Important for backend
            websockets). If None, disables keepalive entirely.
        :type ping_interval: float, optional
        :param ping_timeout: How long to wait for a timeout of a specific ping
            (important for backend websockets.
            Increase only if timeouts occur during runtime).
            If None pings will never time out.
        :type ping_timeout: float, optional
        :param challenge_timeout: How long to wait for agents to challenge.
            If None agent challenging will never time out.
        :type challenge_timeout: float, optional
        :param team: The team to use for formats requiring a team. Can be a showdown
            team string, a showdown packed team string, or a Teambuilder object.
            Defaults to None.
        :type team: str or Teambuilder, optional
        :param choose_on_teampreview: Controls switch-action-based team preview
            selection in formats that support it. If True, team preview uses
            environment actions to pick leads. If False, team preview defaults to
            a random order. If None, it behaves as False (with a warning).
        :type choose_on_teampreview: bool | None
        :param fake: If true, action-order converters will try to avoid returning a default
            output if at all possible, even if the output isn't a legal decision. Defaults
            to False.
        :type fake: bool
        :param strict: If true, action-order converters will throw an error if the move is
            illegal. Otherwise, it will return default. Defaults to True.
        :type strict: bool
        """
        self.metadata = {"name": "poke-env-v0", "render_modes": ["human"]}
        self.render_mode: str | None = None
        self._avatar = avatar
        self._battle_format = battle_format
        self._log_level = log_level
        self._save_replays = save_replays
        self._server_configuration = server_configuration
        self._accept_open_team_sheet = accept_open_team_sheet
        self._start_timer_on_battle_start = start_timer_on_battle_start
        self._start_listening = start_listening
        self._open_timeout = open_timeout
        self._ping_interval = ping_interval
        self._ping_timeout = ping_timeout
        self._challenge_timeout = challenge_timeout
        self._team = team
        self._choose_on_teampreview = choose_on_teampreview
        self._fake = fake
        self._strict = strict
        self._loop = asyncio.new_event_loop()
        Thread(target=self._loop.run_forever, daemon=True).start()
        self.agent1 = _EnvPlayer(
            account_configuration=account_configuration1
            or AccountConfiguration.generate(self.__class__.__name__, rand=True),
            avatar=avatar,
            battle_format=battle_format,
            log_level=log_level,
            max_concurrent_battles=1,
            save_replays=save_replays,
            server_configuration=server_configuration,
            accept_open_team_sheet=accept_open_team_sheet,
            start_timer_on_battle_start=start_timer_on_battle_start,
            start_listening=start_listening,
            open_timeout=open_timeout,
            ping_interval=ping_interval,
            ping_timeout=ping_timeout,
            loop=self._loop,
            team=team,
            choose_on_teampreview=choose_on_teampreview,
        )
        self.agent2 = _EnvPlayer(
            account_configuration=account_configuration2
            or AccountConfiguration.generate(self.__class__.__name__, rand=True),
            avatar=avatar,
            battle_format=battle_format,
            log_level=log_level,
            max_concurrent_battles=1,
            save_replays=save_replays,
            server_configuration=server_configuration,
            accept_open_team_sheet=accept_open_team_sheet,
            start_timer_on_battle_start=start_timer_on_battle_start,
            start_listening=start_listening,
            open_timeout=open_timeout,
            ping_interval=ping_interval,
            ping_timeout=ping_timeout,
            loop=self._loop,
            team=team,
            choose_on_teampreview=choose_on_teampreview,
        )
        self.agents: List[str] = []
        self.possible_agents = [self.agent1.username, self.agent2.username]
        self.battle1: Optional[AbstractBattle] = None
        self.battle2: Optional[AbstractBattle] = None
        self.agent1_to_move = False
        self.agent2_to_move = False
        self._np_random: Optional[Generator] = None
        self._reward_buffer: WeakKeyDictionary[AbstractBattle, float] = (
            WeakKeyDictionary()
        )
        self._challenge_task: Optional[Future[Any]] = None

    def __setattr__(self, name: str, value: Any):
        if name == "observation_spaces":
            value = {
                agent: spaces.Dict(
                    {
                        "observation": raw,
                        "action_mask": spaces.Box(
                            low=0,
                            high=1,
                            shape=(flatdim(self.action_spaces[agent]),),
                            dtype=np.int64,
                        ),
                    }
                )
                for agent, raw in value.items()
            }
        super().__setattr__(name, value)

    def __getstate__(self) -> Dict[str, Any]:
        state = self.__dict__.copy()
        state["_loop"] = None
        state["agent1"] = None
        state["agent2"] = None
        state["_reward_buffer"] = None
        state["_challenge_task"] = None
        return state

    def __setstate__(self, state: Dict[str, Any]):
        self.__dict__.update(state)
        self._loop = asyncio.new_event_loop()
        Thread(target=self._loop.run_forever, daemon=True).start()
        self.agent1 = _EnvPlayer(
            account_configuration=AccountConfiguration.generate(
                self.__class__.__name__, rand=True
            ),
            avatar=self._avatar,
            battle_format=self._battle_format,
            log_level=self._log_level,
            max_concurrent_battles=1,
            save_replays=self._save_replays,
            server_configuration=self._server_configuration,
            accept_open_team_sheet=self._accept_open_team_sheet,
            start_timer_on_battle_start=self._start_timer_on_battle_start,
            start_listening=self._start_listening,
            open_timeout=self._open_timeout,
            ping_interval=self._ping_interval,
            ping_timeout=self._ping_timeout,
            loop=self._loop,
            team=self._team,
            choose_on_teampreview=self._choose_on_teampreview,
        )
        self.agent2 = _EnvPlayer(
            account_configuration=AccountConfiguration.generate(
                self.__class__.__name__, rand=True
            ),
            avatar=self._avatar,
            battle_format=self._battle_format,
            log_level=self._log_level,
            max_concurrent_battles=1,
            save_replays=self._save_replays,
            server_configuration=self._server_configuration,
            accept_open_team_sheet=self._accept_open_team_sheet,
            start_timer_on_battle_start=self._start_timer_on_battle_start,
            start_listening=self._start_listening,
            open_timeout=self._open_timeout,
            ping_interval=self._ping_interval,
            ping_timeout=self._ping_timeout,
            loop=self._loop,
            team=self._team,
            choose_on_teampreview=self._choose_on_teampreview,
        )
        self.agents = []
        old_names = self.possible_agents
        self.possible_agents = [self.agent1.username, self.agent2.username]
        for old, new in zip(old_names, self.possible_agents):
            self.observation_spaces[new] = self.observation_spaces.pop(old)
            self.action_spaces[new] = self.action_spaces.pop(old)
        self._reward_buffer = WeakKeyDictionary()

    ###################################################################################
    # PettingZoo API
    # https://pettingzoo.farama.org/api/parallel/#parallelenv

    def step(
        self, actions: Dict[str, ActionType]
    ) -> Tuple[
        Dict[str, Dict[str, Any]],
        Dict[str, float],
        Dict[str, bool],
        Dict[str, bool],
        Dict[str, Dict[str, Any]],
    ]:
        assert self.battle1 is not None
        assert self.battle2 is not None
        assert not self.battle1.finished
        assert not self.battle2.finished
        if self.agent1_to_move:
            self.agent1_to_move = False
            order1 = self.action_to_order(
                actions[self.agents[0]],
                self.battle1,
                fake=self._fake,
                strict=self._strict,
            )
            self.agent1.order_queue.put(order1)
        if self.agent2_to_move:
            self.agent2_to_move = False
            order2 = self.action_to_order(
                actions[self.agents[1]],
                self.battle2,
                fake=self._fake,
                strict=self._strict,
            )
            self.agent2.order_queue.put(order2)
        battle1 = self.agent1.battle_queue.race_get(
            self.agent1._waiting, self.agent2._trying_again
        )
        battle2 = self.agent2.battle_queue.race_get(
            self.agent2._waiting, self.agent1._trying_again
        )
        self.agent1_to_move = battle1 is not None
        self.agent2_to_move = battle2 is not None
        if battle1 is None:
            self.agent1._waiting.clear()
            self.agent2._trying_again.clear()
            battle1 = self.battle1
        if battle2 is None:
            self.agent2._waiting.clear()
            self.agent1._trying_again.clear()
            battle2 = self.battle2
        observations = {
            self.agents[0]: {
                "observation": self.embed_battle(battle1),
                "action_mask": np.array(self.get_action_mask(battle1)),
            },
            self.agents[1]: {
                "observation": self.embed_battle(battle2),
                "action_mask": np.array(self.get_action_mask(battle2)),
            },
        }
        reward = {
            self.agents[0]: self.calc_reward(battle1),
            self.agents[1]: self.calc_reward(battle2),
        }
        term1, trunc1 = self.calc_term_trunc(battle1)
        term2, trunc2 = self.calc_term_trunc(battle2)
        terminated = {self.agents[0]: term1, self.agents[1]: term2}
        truncated = {self.agents[0]: trunc1, self.agents[1]: trunc2}
        if battle1.finished:
            self.agents = []
        return observations, reward, terminated, truncated, self.get_additional_info()

    def reset(
        self, seed: Optional[int] = None, options: Optional[Dict[str, Any]] = None
    ) -> Tuple[Dict[str, Dict[str, Any]], Dict[str, Dict[str, Any]]]:
        self.agents = [self.agent1.username, self.agent2.username]
        if seed is not None:
            self._np_random, seed = seeding.np_random(seed)
        if self.battle1 and not self.battle1.finished:
            assert self.battle2 is not None
            if self.battle1 == self.agent1.battle:
                if self.agent1_to_move:
                    self.agent1_to_move = False
                    self.agent1.order_queue.put(ForfeitBattleOrder())
                    if self.agent2_to_move:
                        self.agent2_to_move = False
                        self.agent2.order_queue.put(_EmptyBattleOrder())
                else:
                    assert self.agent2_to_move
                    self.agent2_to_move = False
                    self.agent2.order_queue.put(ForfeitBattleOrder())
                self.agent1.battle_queue.get()
                self.agent2.battle_queue.get()
            else:
                raise RuntimeError(
                    "Environment and agent aren't synchronized. Try to restart"
                )
        self.reset_battles()
        self._challenge_task = asyncio.run_coroutine_threadsafe(
            self.agent1.battle_against(self.agent2, n_battles=1), self._loop
        )
        try:
            self.battle1 = self.agent1.battle_queue.get(timeout=self._challenge_timeout)
        except asyncio.TimeoutError:
            raise asyncio.TimeoutError("Agent is not challenging")
        self.battle2 = self.agent2.battle_queue.get()
        self.agent1_to_move = True
        self.agent2_to_move = True
        observations = {
            self.agents[0]: {
                "observation": self.embed_battle(self.battle1),
                "action_mask": np.array(self.get_action_mask(self.battle1)),
            },
            self.agents[1]: {
                "observation": self.embed_battle(self.battle2),
                "action_mask": np.array(self.get_action_mask(self.battle2)),
            },
        }
        return observations, self.get_additional_info()

    def render(self, mode: str = "human"):
        if self.battle1 is not None:
            print(
                "  Turn %4d. | [%s][%3d/%3dhp] %10.10s - %10.10s [%3d%%hp][%s]"
                % (
                    self.battle1.turn,
                    "".join(
                        [
                            "⦻" if mon.fainted else "●"
                            for mon in self.battle1.team.values()
                        ]
                    ),
                    self.battle1.active_pokemon.current_hp or 0,
                    self.battle1.active_pokemon.max_hp or 0,
                    self.battle1.active_pokemon.species,
                    self.battle1.opponent_active_pokemon.species,
                    self.battle1.opponent_active_pokemon.current_hp or 0,
                    "".join(
                        [
                            "⦻" if mon.fainted else "●"
                            for mon in self.battle1.opponent_team.values()
                        ]
                    ),
                ),
                end="\n" if self.battle1.finished else "\r",
            )

    def close(self, force: bool = True, wait: bool = True):
        if force:
            if self.battle1 and not self.battle1.finished:
                assert self.battle2 is not None
                if not self.agent1.battle_queue.empty():
                    self.agent1.battle_queue.get()
                if not self.agent2.battle_queue.empty():
                    self.agent2.battle_queue.get()
                if self.agent1_to_move:
                    self.agent1_to_move = False
                    self.agent1.order_queue.put(ForfeitBattleOrder())
                    if self.agent2_to_move:
                        self.agent2_to_move = False
                        self.agent2.order_queue.put(_EmptyBattleOrder())
                else:
                    assert self.agent2_to_move
                    self.agent2_to_move = False
                    self.agent2.order_queue.put(ForfeitBattleOrder())
        if wait and self._challenge_task is not None:
            self._challenge_task.result()
        if self._challenge_task is None or self._challenge_task.done():
            self.reset_battles()
        self._challenge_task = None
        self.battle1 = None
        self.battle2 = None
        self.agent1.battle = None
        self.agent2.battle = None
        while not self.agent1.order_queue.empty():
            self.agent1.order_queue.get()
        while not self.agent2.order_queue.empty():
            self.agent2.order_queue.get()
        while not self.agent1.battle_queue.empty():
            self.agent1.battle_queue.get()
        while not self.agent2.battle_queue.empty():
            self.agent2.battle_queue.get()

    def observation_space(self, agent: str) -> Space[Dict[str, Any]]:
        return self.observation_spaces[agent]

    def action_space(self, agent: str) -> Space[ActionType]:
        return self.action_spaces[agent]

    ###################################################################################
    # Abstract methods

    @abstractmethod
    def calc_reward(self, battle: AbstractBattle) -> float:
        """
        Returns the reward for the current battle state.

        :param battle: The current battle state.
        :type battle: AbstractBattle

        :return: The reward for battle.
        :rtype: float
        """
        pass

    @abstractmethod
    def embed_battle(self, battle: AbstractBattle) -> Any:
        """
        Returns the embedding of the current battle state in a format compatible with
        the Gymnasium API.

        :param battle: The current battle state.
        :type battle: AbstractBattle

        :return: The embedding of the current battle state.
        """
        pass

    @staticmethod
    @abstractmethod
    def action_to_order(
        action: ActionType, battle: Any, fake: bool = False, strict: bool = True
    ) -> BattleOrder:
        """
        Returns the BattleOrder relative to the given action.

        :param action: The action to take.
        :type action: ActionType
        :param battle: The current battle state
        :type battle: AbstractBattle
        :param fake: If true, action-order converters will try to avoid returning a default
            output if at all possible, even if the output isn't a legal decision. Defaults
            to False.
        :type fake: bool
        :param strict: If true, action-order converters will throw an error if the move is
            illegal. Otherwise, it will return default. Defaults to True.
        :type strict: bool

        :return: The battle order for the given action in context of the current battle.
        :rtype: BattleOrder
        """
        pass

    @staticmethod
    @abstractmethod
    def order_to_action(
        order: BattleOrder, battle: Any, fake: bool = False, strict: bool = True
    ) -> ActionType:
        """
        Returns the action relative to the given BattleOrder.

        :param order: The order to take.
        :type order: BattleOrder
        :param battle: The current battle state
        :type battle: AbstractBattle
        :param fake: If true, action-order converters will try to avoid returning a default
            output if at all possible, even if the output isn't a legal decision. Defaults
            to False.
        :type fake: bool
        :param strict: If true, action-order converters will throw an error if the move is
            illegal. Otherwise, it will return default. Defaults to True.
        :type strict: bool

        :return: The action for the given battle order in context of the current battle.
        :rtype: ActionType
        """
        pass

    @staticmethod
    @abstractmethod
    def get_action_mask(battle: Any) -> List[int]:
        pass

    @staticmethod
    @abstractmethod
    def get_action_space_size(gen: int) -> int:
        pass

    ###################################################################################
    # Helper methods

    def reward_computing_helper(
        self,
        battle: AbstractBattle,
        *,
        fainted_value: float = 0.0,
        hp_value: float = 0.0,
        number_of_pokemons: int = 6,
        starting_value: float = 0.0,
        status_value: float = 0.0,
        victory_value: float = 1.0,
    ) -> float:
        """A helper function to compute rewards.

        The reward is computed by computing the value of a game state, and by comparing
        it to the last state.

        State values are computed by weighting different factor. Fainted pokemons,
        their remaining HP, inflicted statuses and winning are taken into account.

        For instance, if the last time this function was called for battle A it had
        a state value of 8 and this call leads to a value of 9, the returned reward will
        be 9 - 8 = 1.

        Consider a single battle where each player has 6 pokemons. No opponent pokemon
        has fainted, but our team has one fainted pokemon. Three opposing pokemons are
        burned. We have one pokemon missing half of its HP, and our fainted pokemon has
        no HP left.

        The value of this state will be:

        - With fainted value: 1, status value: 0.5, hp value: 1:
            = - 1 (fainted) + 3 * 0.5 (status) - 1.5 (our hp) = -1
        - With fainted value: 3, status value: 0, hp value: 1:
            = - 3 + 3 * 0 - 1.5 = -4.5

        :param battle: The battle for which to compute rewards.
        :type battle: AbstractBattle
        :param fainted_value: The reward weight for fainted pokemons. Defaults to 0.
        :type fainted_value: float
        :param hp_value: The reward weight for hp per pokemon. Defaults to 0.
        :type hp_value: float
        :param number_of_pokemons: The number of pokemons per team. Defaults to 6.
        :type number_of_pokemons: int
        :param starting_value: The default reference value evaluation. Defaults to 0.
        :type starting_value: float
        :param status_value: The reward value per non-fainted status. Defaults to 0.
        :type status_value: float
        :param victory_value: The reward value for winning. Defaults to 1.
        :type victory_value: float
        :return: The reward.
        :rtype: float
        """
        if battle not in self._reward_buffer:
            self._reward_buffer[battle] = starting_value
        current_value = 0.0

        for mon in battle.team.values():
            current_value += mon.current_hp_fraction * hp_value
            if mon.fainted:
                current_value -= fainted_value
            elif mon.status is not None:
                current_value -= status_value

        current_value += (number_of_pokemons - len(battle.team)) * hp_value

        for mon in battle.opponent_team.values():
            current_value -= mon.current_hp_fraction * hp_value
            if mon.fainted:
                current_value += fainted_value
            elif mon.status is not None:
                current_value += status_value

        current_value -= (number_of_pokemons - len(battle.opponent_team)) * hp_value

        if battle.won:
            current_value += victory_value
        elif battle.lost:
            current_value -= victory_value

        to_return = current_value - self._reward_buffer[battle]
        self._reward_buffer[battle] = current_value

        return to_return

    def get_additional_info(self) -> Dict[str, Dict[str, Any]]:
        """
        Returns additional info for the reset method.
        Override only if you really need it.

        :return: Additional information as a Dict
        :rtype: Dict
        """
        return {self.possible_agents[0]: {}, self.possible_agents[1]: {}}

    @staticmethod
    def calc_term_trunc(battle: AbstractBattle):
        terminated = False
        truncated = False
        if battle.finished:
            size = battle.team_size
            remaining_mons = size - len(
                [mon for mon in battle.team.values() if mon.fainted]
            )
            remaining_opponent_mons = size - len(
                [mon for mon in battle.opponent_team.values() if mon.fainted]
            )
            if (remaining_mons == 0) != (remaining_opponent_mons == 0):
                terminated = True
            else:
                truncated = True
        return terminated, truncated

    def reset_battles(self):
        """Resets the player's inner battle tracker."""
        self.agent1.reset_battles()
        self.agent2.reset_battles()

    def done(self, timeout: Optional[int] = None) -> bool:
        """
        Returns True if the task is done or is done after the timeout, false otherwise.

        :param timeout: The amount of time to wait for if the task is not already done.
            If empty it will wait until the task is done.
        :type timeout: int, optional

        :return: True if the task is done or if the task gets completed after the
            timeout.
        :rtype: bool
        """
        if self._challenge_task is None:
            return True
        if timeout is None:
            self._challenge_task.result()
            return True
        if self._challenge_task.done():
            return True
        time.sleep(timeout)
        return self._challenge_task.done()



"""This module defines a base class for players."""

from __future__ import annotations

import asyncio
import random
from abc import ABC, abstractmethod
from asyncio import Condition, Event, Queue, Semaphore
from logging import Logger
from pathlib import Path
from time import perf_counter
from typing import Any, Awaitable, Dict, List, Optional, Union

import orjson

from poke_env.battle.abstract_battle import AbstractBattle
from poke_env.battle.battle import Battle
from poke_env.battle.double_battle import DoubleBattle
from poke_env.battle.move import Move
from poke_env.battle.pokemon import Pokemon
from poke_env.concurrency import (
    POKE_LOOP,
    create_in_poke_loop,
    handle_threaded_coroutines,
)
from poke_env.data import GenData, to_id_str
from poke_env.exceptions import ShowdownException
from poke_env.player.battle_order import (
    BattleOrder,
    DefaultBattleOrder,
    DoubleBattleOrder,
    SingleBattleOrder,
)
from poke_env.ps_client import PSClient
from poke_env.ps_client.account_configuration import AccountConfiguration
from poke_env.ps_client.server_configuration import (
    LocalhostServerConfiguration,
    ServerConfiguration,
)
from poke_env.teambuilder.constant_teambuilder import ConstantTeambuilder
from poke_env.teambuilder.teambuilder import Teambuilder


class Player(ABC):
    """
    Base class for players.
    """

    MESSAGES_TO_IGNORE = {"t:", "expire", "uhtmlchange"}

    # When an error resulting from an invalid choice is made, the next order has this
    # chance of being showdown's default order to prevent infinite loops
    DEFAULT_CHOICE_CHANCE = 1 / 1000

    def __init__(
        self,
        account_configuration: Optional[AccountConfiguration] = None,
        *,
        avatar: Optional[str] = None,
        battle_format: str = "gen9randombattle",
        log_level: Optional[int] = None,
        max_concurrent_battles: int = 1,
        accept_open_team_sheet: bool = False,
        save_replays: Union[bool, str] = False,
        server_configuration: ServerConfiguration = LocalhostServerConfiguration,
        start_timer_on_battle_start: bool = False,
        start_listening: bool = True,
        open_timeout: Optional[float] = 10.0,
        ping_interval: Optional[float] = 20.0,
        ping_timeout: Optional[float] = 20.0,
        loop: asyncio.AbstractEventLoop = POKE_LOOP,
        team: Optional[Union[str, Teambuilder]] = None,
        strict_battle_tracking: bool = False,
    ):
        """
        :param account_configuration: Player configuration. If empty, defaults to an
            automatically generated username with no password. This option must be set
            if the server configuration requires authentication.
        :type account_configuration: AccountConfiguration, optional
        :param avatar: Player avatar name. Optional.
        :type avatar: str, optional
        :param battle_format: Name of the battle format this player plays. Defaults to
            gen9randombattle.
        :type battle_format: str
        :param log_level: The player's logger level.
        :type log_level: int. Defaults to logging's default level.
        :param max_concurrent_battles: Maximum number of battles this player will play
            concurrently. If 0, no limit will be applied. Defaults to 1.
        :type max_concurrent_battles: int
        :param accept_open_team_sheet: Boolean to define whether we want to accept or reject open team
            sheet requests
        :type accept_open_team_sheet: bool
        :param save_replays: Whether to save battle replays. Can be a boolean, where
            True will lead to replays being saved in a potentially new /replay folder,
            or a string representing a folder where replays will be saved.
        :type save_replays: bool or str
        :param server_configuration: Server configuration. Defaults to Localhost Server
            Configuration.
        :type server_configuration: ServerConfiguration
        :param start_listening: Whether to start listening to the server. Defaults to
            True.
        :type start_listening: bool
        :param open_timeout: How long to wait for a timeout when connecting the socket
            (important for backend websockets.
            Increase only if timeouts occur during runtime).
            If None connect will never time out.
        :type open_timeout: float, optional
        :param ping_interval: How long between keepalive pings (Important for backend
            websockets). If None, disables keepalive entirely.
        :type ping_interval: float, optional
        :param ping_timeout: How long to wait for a timeout of a specific ping
            (important for backend websockets.
            Increase only if timeouts occur during runtime).
            If None pings will never time out.
        :type ping_timeout: float, optional
        :param start_timer_on_battle_start: Whether to automatically start the battle
            timer on battle start. Defaults to False.
        :type start_timer_on_battle_start: bool
        :param team: The team to use for formats requiring a team. Can be a showdown
            team string, a showdown packed team string, or a Teambuilder object.
            Defaults to None.
        :type team: str or Teambuilder, optional
        """
        self.ps_client = PSClient(
            account_configuration=account_configuration
            or AccountConfiguration.generate(self.__class__.__name__),
            avatar=avatar,
            log_level=log_level,
            server_configuration=server_configuration,
            start_listening=start_listening,
            open_timeout=open_timeout,
            ping_interval=ping_interval,
            ping_timeout=ping_timeout,
            loop=loop,
        )

        self.ps_client._handle_battle_message = self._handle_battle_message  # type: ignore
        self.ps_client._update_challenges = self._update_challenges  # type: ignore
        self.ps_client._handle_challenge_request = self._handle_challenge_request  # type: ignore

        self._format: str = battle_format
        self._max_concurrent_battles: int = max_concurrent_battles
        self._save_replays = save_replays
        self._start_timer_on_battle_start: bool = start_timer_on_battle_start
        self._accept_open_team_sheet: bool = accept_open_team_sheet

        self._battles: Dict[str, AbstractBattle] = {}
        self._battle_semaphore: Semaphore = create_in_poke_loop(Semaphore, loop, 0)

        self._battle_start_condition: Condition = create_in_poke_loop(Condition, loop)
        self._battle_count_queue: Queue[Any] = create_in_poke_loop(
            Queue, loop, max_concurrent_battles
        )
        self._battle_end_condition: Condition = create_in_poke_loop(Condition, loop)
        self._challenge_queue: Queue[Any] = create_in_poke_loop(Queue, loop)
        self._waiting: Event = create_in_poke_loop(Event, loop)
        self._trying_again: Event = create_in_poke_loop(Event, loop)
        self._team: Optional[Teambuilder] = None
        self._current_packed_team: str | None = None
        self._strict_battle_tracking = strict_battle_tracking

        if isinstance(team, Teambuilder):
            self._team = team
        elif isinstance(team, str):
            self._team = ConstantTeambuilder(team)

        self.logger.debug("Player initialisation finished")

    def _battle_finished_callback(self, battle: AbstractBattle):
        pass

    def update_team(self, team: Union[Teambuilder, str]):
        """Updates the team used by the player.

        :param team: The new team to use.
        :type team: str or Teambuilder
        """
        if isinstance(team, Teambuilder):
            self._team = team
        else:
            self._team = ConstantTeambuilder(team)

    async def _create_battle(self, split_message: List[str]) -> AbstractBattle:
        """Returns battle object corresponding to received message.

        :param split_message: The battle initialisation message.
        :type split_message: List[str]
        :return: The corresponding battle object.
        :rtype: AbstractBattle
        """
        # We check that the battle has the correct format
        if split_message[1] == self._format and len(split_message) >= 2:
            # Battle initialisation
            battle_tag = "-".join(split_message)[1:]

            if battle_tag in self._battles:
                return self._battles[battle_tag]
            else:
                gen = GenData.from_format(self._format).gen
                if self.format_is_doubles:
                    battle: AbstractBattle = DoubleBattle(
                        battle_tag=battle_tag,
                        username=self.username,
                        logger=self.logger,
                        save_replays=self._save_replays,
                        gen=gen,
                    )
                else:
                    battle = Battle(
                        battle_tag=battle_tag,
                        username=self.username,
                        logger=self.logger,
                        gen=gen,
                        save_replays=self._save_replays,
                    )

                if self._current_packed_team:
                    battle._teambuilder_team = Teambuilder.parse_packed_team(
                        self._current_packed_team
                    )

                await self._battle_count_queue.put(None)
                if battle_tag in self._battles:
                    await self._battle_count_queue.get()
                    return self._battles[battle_tag]
                async with self._battle_start_condition:
                    self._battle_semaphore.release()
                    self._battle_start_condition.notify_all()
                    self._battles[battle_tag] = battle

                if self._start_timer_on_battle_start:
                    await self.ps_client.send_message("/timer on", battle.battle_tag)

                if hasattr(self.ps_client, "websocket") and "vgc" in self.format:
                    if self.accept_open_team_sheet:
                        await self.ps_client.send_message(
                            "/acceptopenteamsheets", room=battle_tag
                        )
                    else:
                        await self.ps_client.send_message(
                            "/rejectopenteamsheets", room=battle_tag
                        )

                return battle
        else:
            self.logger.critical(
                "Unmanaged battle initialisation message received: %s", split_message
            )
            raise ShowdownException()

    async def _get_battle(self, battle_tag: str) -> AbstractBattle:
        battle_tag = battle_tag[1:]
        while True:
            if battle_tag in self._battles:
                return self._battles[battle_tag]
            async with self._battle_start_condition:
                await self._battle_start_condition.wait()

    async def _handle_battle_message(self, split_messages: List[List[str]]):
        """Handles a battle message.

        :param split_message: The received battle message.
        :type split_message: str
        """
        # Battle messages can be multiline
        if (
            len(split_messages) > 1
            and len(split_messages[1]) > 1
            and split_messages[1][1] == "init"
        ):
            battle_info = split_messages[0][0].split("-")
            battle = await self._create_battle(battle_info)
        else:
            battle = await self._get_battle(split_messages[0][0])

        for split_message in split_messages[1:]:
            if not split_message:
                continue
            elif len(split_message) == 1:
                if (
                    battle.teampreview
                    and self.accept_open_team_sheet
                    and "rejected open team sheets." in split_message[0]
                ):
                    await self._handle_battle_request(battle)
            elif split_message[1] == "":
                battle.parse_message(split_message)
            elif split_message[1] in self.MESSAGES_TO_IGNORE:
                pass
            elif split_message[1] == "request":
                if split_message[2]:
                    request = orjson.loads(split_message[2])
                    battle.parse_request(request, self._strict_battle_tracking)
                    if not (battle.teampreview and self.accept_open_team_sheet):
                        await self._handle_battle_request(battle)
            elif split_message[1] == "showteam":
                role = split_message[2]
                if role == battle.player_role:
                    continue
                teambuilder_team = Teambuilder.parse_packed_team(
                    "|".join(split_message[3:])
                )
                battle.apply_teambuilder_team(
                    role, teambuilder_team, battle.teampreview_opponent_team
                )
                await self._handle_battle_request(battle)
            elif split_message[1] == "win" or split_message[1] == "tie":
                if split_message[1] == "win":
                    battle.won_by(split_message[2])
                else:
                    battle.tied()
                await self._battle_count_queue.get()
                self._battle_count_queue.task_done()
                self._battle_finished_callback(battle)
                async with self._battle_end_condition:
                    self._battle_end_condition.notify_all()
                self.ps_client._battle_locks.pop(battle.battle_tag)
                if hasattr(self.ps_client, "websocket"):
                    await self.ps_client.send_message(f"/leave {battle.battle_tag}")
            elif split_message[1] == "error":
                self.logger.log(
                    25, "Error message received: %s", "|".join(split_message)
                )
                if split_message[2].startswith("[Unavailable choice]"):
                    self._trying_again.set()
                elif split_message[2].startswith("[Invalid choice]"):
                    await self._handle_battle_request(battle, maybe_default_order=True)
                else:
                    self.logger.critical("Unexpected error message: %s", split_message)
            elif split_message[1] == "bigerror":
                self.logger.warning("Received 'bigerror' message: %s", split_message)
            else:
                battle.parse_message(split_message)

    async def _handle_battle_request(
        self, battle: AbstractBattle, maybe_default_order: bool = False
    ):
        if battle._wait:
            self._waiting.set()
            return
        if maybe_default_order and random.random() < self.DEFAULT_CHOICE_CHANCE:
            message = self.choose_default_move().message
        elif battle.teampreview:
            m = self.teampreview(battle)
            if isinstance(m, Awaitable):
                m = await m
            message = m
        else:
            if maybe_default_order:
                self._trying_again.set()
            choice = self.choose_move(battle)
            if isinstance(choice, Awaitable):
                choice = await choice
            message = choice.message
        if message:
            await self.ps_client.send_message(message, battle.battle_tag)

    async def _handle_challenge_request(self, split_message: List[str]):
        """Handles an individual challenge."""
        challenging_player = split_message[2].strip()

        if challenging_player != self.username:
            if len(split_message) >= 6:
                if split_message[5] == self._format:
                    await self._challenge_queue.put(challenging_player)

    async def _update_challenges(self, split_message: List[str]):
        """Update internal challenge state.

        Add corresponding challenges to internal queue of challenges, where they will be
        processed if relevant.

        :param split_message: Recevied message, split.
        :type split_message: List[str]
        """
        self.logger.debug("Updating challenges with %s", split_message)
        challenges = orjson.loads(split_message[2]).get("challengesFrom", {})
        for user, format_ in challenges.items():
            if format_ == self._format:
                await self._challenge_queue.put(user)

    async def accept_challenges(
        self,
        opponent: Optional[Union[str, List[str]]],
        n_challenges: int,
        packed_team: Optional[str] = None,
    ):
        """Let the player wait for challenges from opponent, and accept them.

        If opponent is None, every challenge will be accepted. If opponent if a string,
        all challenges from player with that name will be accepted. If opponent is a
        list all challenges originating from players whose name is in the list will be
        accepted.

        Up to n_challenges challenges will be accepted, after what the function will
        wait for these battles to finish, and then return.

        :param opponent: Players from which challenges will be accepted.
        :type opponent: None, str or list of str
        :param n_challenges: Number of challenges that will be accepted
        :type n_challenges: int
        :packed_team: Team to use. Defaults to generating a team with the agent's teambuilder.
        :type packed_team: string, optional.
        """
        await handle_threaded_coroutines(
            self._accept_challenges(opponent, n_challenges, packed_team),
            self.ps_client.loop,
        )

    async def _accept_challenges(
        self,
        opponent: Optional[Union[str, List[str]]],
        n_challenges: int,
        packed_team: Optional[str],
    ):
        if opponent:
            if isinstance(opponent, list):
                opponent = [to_id_str(o) for o in opponent]
            else:
                opponent = to_id_str(opponent)
        await self.ps_client.logged_in.wait()
        self.logger.debug("Event logged in received in accept_challenge")

        for _ in range(n_challenges):
            if packed_team:
                self._current_packed_team = packed_team
            else:
                self.get_next_team()
            while True:
                username = to_id_str(await self._challenge_queue.get())
                self.logger.debug(
                    "Consumed %s from challenge queue in accept_challenge", username
                )
                if (
                    (opponent is None)
                    or (opponent == username)
                    or (isinstance(opponent, list) and (username in opponent))
                ):
                    await self.ps_client.accept_challenge(
                        username, self._current_packed_team
                    )
                    await self._battle_semaphore.acquire()
                    break
        await self._battle_count_queue.join()

    @abstractmethod
    def choose_move(
        self, battle: AbstractBattle
    ) -> Union[BattleOrder, Awaitable[BattleOrder]]:
        """Abstract method to choose a move in a battle.

        :param battle: The battle.
        :type battle: AbstractBattle
        :return: The move order, or an awaitable resolving to one.
        :rtype: BattleOrder or Awaitable[BattleOrder]
        """
        pass

    @staticmethod
    def choose_default_move() -> DefaultBattleOrder:
        """Returns showdown's default move order.

        This order will result in the first legal order - according to showdown's
        ordering - being chosen.
        """
        return DefaultBattleOrder()

    @staticmethod
    def choose_random_doubles_move(battle: DoubleBattle) -> DoubleBattleOrder:
        orders = DoubleBattleOrder.join_orders(*battle.valid_orders)
        if orders:
            return orders[int(random.random() * len(orders))]
        else:
            return DoubleBattleOrder(DefaultBattleOrder(), DefaultBattleOrder())

    @staticmethod
    def choose_random_singles_move(battle: Battle) -> SingleBattleOrder:
        orders = battle.valid_orders
        if orders:
            return orders[int(random.random() * len(orders))]
        else:
            return Player.choose_default_move()

    @staticmethod
    def choose_random_move(battle: AbstractBattle) -> BattleOrder:
        """Returns a random legal move from battle.

        :param battle: The battle in which to move.
        :type battle: AbstractBattle
        :return: Move order
        :rtype: str
        """
        if isinstance(battle, DoubleBattle):
            return Player.choose_random_doubles_move(battle)
        elif isinstance(battle, Battle):
            return Player.choose_random_singles_move(battle)
        else:
            raise ValueError(
                f"battle should be Battle or DoubleBattle. Received {type(battle)}"
            )

    async def ladder(self, n_games: int):
        """Make the player play games on the ladder.

        n_games defines how many battles will be played.

        :param n_games: Number of battles that will be played
        :type n_games: int
        """
        await handle_threaded_coroutines(self._ladder(n_games), self.ps_client.loop)

    async def _ladder(self, n_games: int):
        await self.ps_client.logged_in.wait()
        start_time = perf_counter()

        for _ in range(n_games):
            async with self._battle_start_condition:
                await self.ps_client.search_ladder_game(
                    self._format, self.get_next_team()
                )
                await self._battle_start_condition.wait()
                while self._battle_count_queue.full():
                    async with self._battle_end_condition:
                        await self._battle_end_condition.wait()
                await self._battle_semaphore.acquire()
        await self._battle_count_queue.join()
        self.logger.info(
            "Laddering (%d battles) finished in %fs",
            n_games,
            perf_counter() - start_time,
        )

    async def battle_against(self, *opponents: Player, n_battles: int = 1):
        """Make the player play n_battles against the given opponents.

        This function is a wrapper around send_challenges and accept_challenges.

        :param opponents: The opponents to play against.
        :type opponents: Player
        :param n_battles: The number of games to play. Defaults to 1.
        :type n_battles: int
        """
        await handle_threaded_coroutines(
            self._battle_against(*opponents, n_battles=n_battles), self.ps_client.loop
        )

    async def _battle_against(self, *opponents: Player, n_battles: int):
        for opponent in opponents:
            await asyncio.gather(
                self.send_challenges(
                    to_id_str(opponent.username),
                    n_battles,
                    to_wait=opponent.ps_client.logged_in,
                ),
                opponent.accept_challenges(to_id_str(self.username), n_battles),
            )

    async def send_challenges(
        self, opponent: str, n_challenges: int, to_wait: Optional[Event] = None
    ):
        """Make the player send challenges to opponent.

        opponent must be a string, corresponding to the name of the player to challenge.

        n_challenges defines how many challenges will be sent.

        to_wait is an optional event that can be set, in which case it will be waited
        before launching challenges.

        :param opponent: Player username to challenge.
        :type opponent: str
        :param n_challenges: Number of battles that will be started
        :type n_challenges: int
        :param to_wait: Optional event to wait before launching challenges.
        :type to_wait: Event, optional.
        """
        await handle_threaded_coroutines(
            self._send_challenges(opponent, n_challenges, to_wait), self.ps_client.loop
        )

    async def _send_challenges(
        self, opponent: str, n_challenges: int, to_wait: Optional[Event] = None
    ):
        await self.ps_client.logged_in.wait()
        self.logger.info("Event logged in received in send challenge")

        if to_wait is not None:
            await to_wait.wait()

        start_time = perf_counter()

        for _ in range(n_challenges):
            await self.ps_client.challenge(opponent, self._format, self.get_next_team())
            await self._battle_semaphore.acquire()
        await self._battle_count_queue.join()
        self.logger.info(
            "Challenges (%d battles) finished in %fs",
            n_challenges,
            perf_counter() - start_time,
        )

    def random_teampreview(self, battle: AbstractBattle) -> str:
        """Returns a random valid teampreview order for the given battle.

        :param battle: The battle.
        :type battle: AbstractBattle
        :return: The random teampreview order.
        :rtype: str
        """
        members = list(range(1, len(battle.team) + 1))
        random.shuffle(members)
        if battle.format is not None and "vgc" in battle.format:
            members = members[:4]
        for i in members:
            list(battle.team.values())[i - 1]._selected_in_teampreview = True
        return "/team " + "".join([str(c) for c in members])

    def reset_battles(self):
        """Resets the player's inner battle tracker."""
        for battle in list(self._battles.values()):
            if not battle.finished:
                raise EnvironmentError(
                    "Can not reset player's battles while they are still running"
                )
        self._battles = {}

    def teampreview(self, battle: AbstractBattle) -> Union[str, Awaitable[str]]:
        """Returns a teampreview order for the given battle.

        This order must be of the form /team TEAM, where TEAM is a string defining the
        team chosen by the player. Multiple formats are supported, among which '3461'
        and '3, 4, 6, 1' are correct and indicate leading with pokemon 3, with pokemons
        4, 6 and 1 in the back in single battles or leading with pokemons 3 and 4 with
        pokemons 6 and 1 in the back in double battles.

        Please refer to Pokemon Showdown's protocol documentation for more information.

        IMPORTANT: If overwriting this method, make sure to mark Pokemon selected in
        teampreview as such by setting their _selected_in_teampreview attribute to True.
        See random_teampreview implementation for an example.

        :param battle: The battle.
        :type battle: AbstractBattle
        :return: The teampreview order.
        :rtype: str
        """
        return self.random_teampreview(battle)

    @staticmethod
    def create_order(
        order: Union[Move, Pokemon],
        mega: bool = False,
        z_move: bool = False,
        dynamax: bool = False,
        terastallize: bool = False,
        move_target: int = DoubleBattle.EMPTY_TARGET_POSITION,
    ) -> SingleBattleOrder:
        """Formats an move order corresponding to the provided pokemon or move.

        :param order: Move to make or Pokemon to switch to.
        :type order: Move or Pokemon
        :param mega: Whether to mega evolve the pokemon, if a move is chosen.
        :type mega: bool
        :param z_move: Whether to make a zmove, if a move is chosen.
        :type z_move: bool
        :param dynamax: Whether to dynamax, if a move is chosen.
        :type dynamax: bool
        :param terastallize: Whether to terastallize, if a move is chosen.
        :type terastallize: bool
        :param move_target: Target Pokemon slot of a given move
        :type move_target: int
        :return: Formatted move order
        :rtype: str
        """
        return SingleBattleOrder(
            order,
            mega=mega,
            move_target=move_target,
            z_move=z_move,
            dynamax=dynamax,
            terastallize=terastallize,
        )

    def save_replay(self, battle_tag: str, file_path: Union[str, Path]) -> Path:
        """
        Saves replay html for a known battle.

        :param battle_tag: The battle tag to export (with or without leading '>').
        :type battle_tag: str
        :param file_path: Path where replay html should be written.
        :type file_path: str | pathlib.Path
        :return: The written replay path.
        :rtype: pathlib.Path
        :raises KeyError: If battle_tag does not correspond to a known battle.
        """
        normalized_battle_tag = (
            battle_tag[1:] if battle_tag.startswith(">") else battle_tag
        )
        battle = self._battles.get(normalized_battle_tag)
        if battle is None:
            known_battle_tags = ", ".join(sorted(self._battles.keys())) or "<none>"
            raise KeyError(
                f"Unknown battle_tag '{battle_tag}'. Known battle tags: {known_battle_tags}"
            )
        return battle.save_replay(file_path)

    def get_next_team(self) -> str | None:
        if self._team:
            self._current_packed_team = self._team.yield_team()
            return self._current_packed_team
        return None

    @property
    def battles(self) -> Dict[str, AbstractBattle]:
        return self._battles

    @property
    def format(self) -> str:
        return self._format

    @property
    def format_is_doubles(self) -> bool:
        format_lowercase = self._format.lower()
        return (
            "vgc" in format_lowercase
            or "double" in format_lowercase
            or "metronome" in format_lowercase
        )

    @property
    def n_finished_battles(self) -> int:
        return len([None for b in self._battles.values() if b.finished])

    @property
    def n_lost_battles(self) -> int:
        return len([None for b in self._battles.values() if b.lost])

    @property
    def n_tied_battles(self) -> int:
        return self.n_finished_battles - self.n_lost_battles - self.n_won_battles

    @property
    def n_won_battles(self) -> int:
        return len([None for b in self._battles.values() if b.won])

    @property
    def accept_open_team_sheet(self) -> bool:
        return self._accept_open_team_sheet

    @property
    def win_rate(self) -> float:
        return self.n_won_battles / self.n_finished_battles

    @property
    def logger(self) -> Logger:
        return self.ps_client.logger

    @property
    def username(self) -> str:
        return self.ps_client.username



"""This module defines a gen9 damage calculator"""

import math
from typing import Dict, List, Optional, Union

from poke_env.battle import (
    Battle,
    DoubleBattle,
    Effect,
    Field,
    Move,
    MoveCategory,
    Pokemon,
    PokemonGender,
    PokemonType,
    SideCondition,
    Status,
    Target,
    Weather,
)
from poke_env.data import GenData


# Source: https://github.com/smogon/damage-calc/blob/master/calc/src/mechanics/gen789.ts#L52
# NOTE: dont deal with multihits (just take average)
# NOTE: don't deal with shell side arm, naturalgift, technoblast, pursuit, multiattack, aurawheel, fling
# NOTE: don't deal with xerneas or zygarde or yveltal auras, or battlebond
# NOTE: dont deal with metronome (the item) or stellar because we dont keep a log of historical move usage
# NOTE: I ignore dynamax, mega, primal,battlebond and parental bond (not legal in gen9)
# NOTE: knockoff damage isnt realllllly right cuz I dont look at pokemon/item pairs, just items
def calculate_damage(
    attacker_identifier: str,
    defender_identifier: str,
    move: Move,
    battle: Union[Battle, DoubleBattle],
    is_critical: bool = False,
):
    """Return the possible damage range for a move.

    :param attacker_identifier: Identifier of the attacking Pokémon.
    :type attacker_identifier: str
    :param defender_identifier: Identifier of the defending Pokémon.
    :type defender_identifier: str
    :param move: Move being used.
    :type move: Move
    :param battle: Current battle object containing both Pokémon.
    :type battle: Battle or DoubleBattle
    :param is_critical: Whether to compute damage for a critical hit.
    :type is_critical: bool
    :return: Tuple of minimum and maximum damage rolls.
    :rtype: Tuple[int, int]

    Notes that several edge cases are ignored and behaviour may deviate from
    the official damage calculator.
    """
    attacker = battle.get_pokemon(attacker_identifier)
    defender = battle.get_pokemon(defender_identifier)

    assert battle.player_role is not None and battle.opponent_role is not None
    assert all(
        map(
            lambda x: isinstance(x, int) or isinstance(x, float),
            attacker.stats.values(),
        )
    ), "attacker stats not defined"
    assert all(
        map(
            lambda x: isinstance(x, int) or isinstance(x, float),
            defender.stats.values(),
        )
    ), "defender stats not defined"

    move_category = move.category
    flags = {x: 1 for x in move.entry.get("flags", {})}

    # NOTE: I don't deal with shell side arm
    if (
        move.id in ["photongeyser", "terablast"]
        and attacker.stats["atk"]
        and attacker.stats["spa"]
    ):
        move_category = (
            MoveCategory.PHYSICAL
            if attacker.stats["atk"] * BOOST_MULTIPLIERS[attacker.boosts["atk"]]
            > attacker.stats["spa"] * BOOST_MULTIPLIERS[attacker.boosts["spa"]]
            else MoveCategory.SPECIAL
        )

    if move_category == MoveCategory.STATUS and not move.id == "naturepower":
        return 0, 0

    if flags.get("punch", 0) and attacker.item == "punchingglove":
        flags["contact"] = 0

    move_breaks_protect = move.breaks_protect or (
        attacker.ability == "unseenfist" and flags.get("contact", 0) == 1
    )

    if (
        any(map(lambda x: x in PROTECT_EFFECTS, defender.effects))
        and not move_breaks_protect
    ):
        return 0, 0

    if move.id == "painsplit":
        attacker_hp_stat = attacker.stats.get("hp", None)
        defender_hp_stat = defender.stats.get("hp", None)
        if attacker_hp_stat is not None and defender_hp_stat is not None:
            attacker_hp = (
                math.floor(attacker_hp_stat * attacker.current_hp_fraction)
                if attacker_hp_stat
                else 0
            )
            defender_hp = (
                math.floor(defender_hp_stat * defender.current_hp_fraction)
                if defender_hp_stat
                else 0
            )
            damage = max(0, defender_hp - math.floor((attacker_hp + defender_hp) / 2))
            return damage, damage
        else:
            return 0, 0

    defender_ability = defender.ability
    attacker_ability = attacker.ability
    if move.id in MOVE_IGNORES_ABILITY or (
        attacker.ability in ATTACKER_IGNORES_ABILITY
        and not defender.item == "abilityshield"
    ):
        defender_ability = ""

    if any(
        map(lambda x: x and x.ability == "neutralizinggas", battle.all_active_pokemons)
    ):
        if (
            defender.ability not in ABILITY_IGNORES_NEUTRALIZING_GAS
            or not defender.item == "abilityshield"
        ):
            defender_ability = ""
        elif (
            attacker.ability not in ABILITY_IGNORES_NEUTRALIZING_GAS
            or not attacker.item == "abilityshield"
        ):
            attacker_ability = ""

    if defender_ability in ["battlearmor", "shellarmor"]:
        is_critical = False
    elif attacker_ability == "merciless" and defender.status in [
        Status.PSN,
        Status.TOX,
    ]:
        is_critical = True

    defender_weight = defender.weight
    if defender_ability == "lightmetal":
        defender_weight *= 0.5
    elif defender_ability == "heavymetal":
        defender_weight *= 2
    if defender.item == "floatstone":
        defender_weight *= 0.5

    # NOTE: no support for technoblast or multiattack or naturalgift, or aurawheel
    move_type: Optional[PokemonType] = move.type
    move_target: Target = move.target if move.target is not None else Target.SELF
    if move.id == "weatherball":
        if Weather.SUNNYDAY in battle.weather and attacker.item != "utilityumbrella":
            move_type = PokemonType.FIRE
        elif Weather.RAINDANCE in battle.weather and attacker.item != "utilityumbrella":
            move_type = PokemonType.WATER
        elif Weather.SANDSTORM in battle.weather:
            move_type = PokemonType.ROCK
        elif Weather.HAIL in battle.weather or Weather.SNOW in battle.weather:
            move_type = PokemonType.ICE
        else:
            move_type = PokemonType.NORMAL
    elif move.id == "judgment" and attacker.item and attacker.item.endswith("plate"):
        move_type = get_item_boost_type(attacker.item)
    elif move.id == "naturepower" or (
        move.id == "terrainpulse" and battle.is_grounded(attacker)
    ):
        if Field.ELECTRIC_TERRAIN in battle.fields:
            move_type = PokemonType.ELECTRIC
        elif Field.GRASSY_TERRAIN in battle.fields:
            move_type = PokemonType.GRASS
        elif Field.MISTY_TERRAIN in battle.fields:
            move_type = PokemonType.FAIRY
        elif Field.PSYCHIC_TERRAIN in battle.fields:
            move_type = PokemonType.PSYCHIC
        else:
            move_type = PokemonType.NORMAL
    elif move.id == "revelationdance":
        move_type = attacker.type_1
    elif move.id == "ragingbull":
        if attacker.species == "taurospaldeacombat":
            move_type = PokemonType.FIGHTING
        elif attacker.species == "taurospaldeablaze":
            move_type = PokemonType.FIRE
        elif attacker.species == "taurospaldeaaqua":
            move_type = PokemonType.WATER

        flags["ignorescreens"] = 1
    elif move.id == "ivycudgel":
        if attacker.species == "ogerponcornerstone":
            move_type = PokemonType.ROCK
        elif attacker.species == "ogerponhearthflame":
            move_type = PokemonType.FIRE
        elif attacker.species == "ogerponwellspring":
            move_type = PokemonType.WATER
    elif move.id == "terastarstorm" and attacker.species == "terapagosstellar":
        move_target = Target.ALL_ADJACENT_FOES
        move_type = PokemonType.STELLAR
    elif move.id in ["brickbreak", "psychicfangs"]:
        flags["ignorescreens"] = 1
    # NOTE: had to add this; not present in showdown file
    elif move.id == "expandingforce" and Field.PSYCHIC_TERRAIN in battle.fields:
        move_target = Target.ALL_ADJACENT_FOES

    has_ate_ability_type_change = False  # measures whether the move type has changed by an -ate ability (eg pixilate)
    no_type_change = move.id in {
        "revelationdance",
        "judgment",
        "naturepower",
        "technoblast",
        "multiattack",
        "naturalgift",
        "weatherball",
        "terrainpulse",
        "struggle",
    } or (move.id == "terablast" and attacker.is_terastallized)

    if not no_type_change:
        if attacker_ability == "aerilate" and move_type == PokemonType.NORMAL:
            move_type = PokemonType.FLYING
            has_ate_ability_type_change = True
        elif attacker_ability == "galvanize" and move_type == PokemonType.NORMAL:
            move_type = PokemonType.ELECTRIC
            has_ate_ability_type_change = True
        elif attacker_ability == "liquidvoice" and flags.get("sound", 0) == 1:
            move_type = PokemonType.WATER
        elif attacker_ability == "pixilate" and move_type == PokemonType.NORMAL:
            move_type = PokemonType.FAIRY
            has_ate_ability_type_change = True
        elif attacker_ability == "refrigerate" and move_type == PokemonType.NORMAL:
            move_type = PokemonType.ICE
            has_ate_ability_type_change = True
        elif attacker_ability == "normalize":
            move_type = PokemonType.NORMAL
            has_ate_ability_type_change = True

    if move.id == "terablast" and attacker.is_terastallized:
        move_type = attacker.type_1

    move_priority = move.priority
    if (attacker_ability == "triage" and "heal" in flags) or (
        attacker_ability == "galewings"
        and move_type == PokemonType.FLYING
        and attacker.current_hp_fraction == 100
    ):
        move_priority = 1

    first_type_effectiveness = get_move_effectiveness(
        move,
        move_type or move.type,
        defender.type_1,
        attacker_ability in ["scrappy", "mindseye"],
        Field.GRAVITY in battle.fields,
        defender.item == "ringtarget",
    )
    second_type_effectiveness = (
        1
        if len(defender.types) == 1
        else get_move_effectiveness(
            move,
            move_type or move.type,
            defender.type_2,  # type: ignore
            attacker_ability in ["scrappy", "mindseye"],
            Field.GRAVITY in battle.fields,
            defender.item == "ringtarget",
        )
    )

    type_effectiveness = first_type_effectiveness * second_type_effectiveness

    if (
        type_effectiveness == 0
        and move_type == PokemonType.GROUND
        and defender.item == "ironball"
    ):
        type_effectiveness = 1

    if type_effectiveness == 0 and move.id == "thousandarrows":
        type_effectiveness = 1

    if type_effectiveness == 0:
        return 0, 0

    if move.id == "skydrop" and (
        PokemonType.FLYING in defender.types
        or defender_weight >= 200
        or Field.GRAVITY in battle.fields
    ):
        return 0, 0
    elif move.id == "synchronoise" and not (
        attacker.type_1 in defender.types or attacker.type_2 in defender.types
    ):
        return 0, 0
    elif move.id == "dreameater" and not (
        defender.status == Status.SLP or defender_ability == "comatose"
    ):
        return 0, 0
    elif move.id == "steelroller" and not any(
        map(lambda x: x.is_terrain, battle.fields)
    ):
        return 0, 0
    elif move.id == "poltergeist" and not defender.item:
        return 0, 0

    if move_type == PokemonType.STELLAR:
        type_effectiveness = 2 if defender.is_terastallized else 1

    if defender_ability == "terashell" and defender.current_hp_fraction == 100:
        type_effectiveness = 0.5

    if defender_ability == "wonderguard" and type_effectiveness <= 1:
        return 0, 0
    elif move_type == PokemonType.GRASS and defender_ability == "sapsipper":
        return 0, 0
    elif move_type == PokemonType.FIRE and defender_ability in [
        "flashfire",
        "wellbakedbody",
    ]:
        return 0, 0
    elif move_type == PokemonType.WATER and defender_ability in [
        "dryskin",
        "stormdrain",
        "waterabsorb",
    ]:
        return 0, 0
    elif move_type == PokemonType.ELECTRIC and defender_ability in [
        "lightningrod",
        "motordrive",
        "voltabsorb",
    ]:
        return 0, 0
    elif (
        move_type == PokemonType.GROUND
        and Field.GRAVITY not in battle.fields
        and not move.id == "thousandarrows"
        and not defender.item == "ironball"
        and defender.ability == "levitate"
    ):
        return 0, 0
    elif flags.get("bullet", 0) == 1 and defender_ability == "bulletproof":
        return 0, 0
    elif (
        flags.get("sound", 0) == 1
        and not move.id == "clangoroussoul"
        and defender_ability == "soundproof"
    ):
        return 0, 0
    elif move.priority > 0 and defender_ability in [
        "queenlymajesty",
        "dazzling",
        "armortail",
    ]:
        return 0, 0
    elif move_type == PokemonType.GROUND and defender_ability == "eartheater":
        return 0, 0
    elif flags.get("wind", 0) == 1 and defender_ability == "windrider":
        return 0, 0
    elif (
        move_type == PokemonType.GROUND
        and move.id != "thousandarrows"
        and Field.GRAVITY not in battle.fields
        and defender.item == "airballoon"
    ):
        return 0, 0
    elif (
        move_priority > 0
        and Field.PSYCHIC_TERRAIN in battle.fields
        and battle.is_grounded(defender)
    ):
        return 0, 0

    if move.id in ["seismictoss", "nightshade"]:
        return attacker.level, attacker.level
    elif move.id == "dragonrage":
        return 40, 40
    elif move.id == "sonicboom":
        return 20, 20

    # NOTE: this will be wrong if we use just Showdown ingested
    if move.id == "finalgambit":
        hp_stat = attacker.stats.get("hp", None)
        damage = math.floor(hp_stat * attacker.current_hp_fraction) if hp_stat else 0
        return damage, damage

    base_power = calculate_base_power(
        attacker_identifier,
        defender_identifier,
        move,
        battle,
        has_ate_ability_type_change,
    )

    # NOTE: I don't want to deal with multihit complexity so I calculate the average damage expected
    # This makes damage calc inaccurate for tripleaxel
    if move.n_hit != (1, 1):
        base_power = base_power * ((move.n_hit[0] + move.n_hit[1]) / 2)

    if base_power == 0:
        return 0, 0

    attack = calculate_attack(
        attacker_identifier, defender_identifier, move, battle, is_critical
    )
    defense = calculate_defense(
        attacker_identifier, defender_identifier, move, battle, is_critical
    )

    base_damage = calculate_base_damage(
        attacker_identifier,
        defender_identifier,
        base_power,
        attack,
        defense,
        move,
        move_type,
        move_target,
        battle,
        is_critical,
    )

    # the random factor is applied between the crit mod and the stab mod, so don't apply anything
    # below this until we're inside the loop
    pre_stellar_stab_mod = get_stab_mod(attacker, move, move_type)
    stab_mod = get_stellar_stab_mod(attacker, move, pre_stellar_stab_mod)

    apply_burn = (
        attacker.status == Status.BRN
        and move_category == MoveCategory.PHYSICAL
        and not attacker_ability == "guts"
        and not move.id == "facade"
    )

    final_mods = calculate_final_mods(
        attacker_identifier,
        defender_identifier,
        move,
        battle,
        type_effectiveness,
        flags,
        is_critical,
    )

    final_mod = chain_mods(final_mods, 41, 131072)

    return (
        get_final_damage(
            base_damage, 0, type_effectiveness, apply_burn, stab_mod, final_mod
        ),
        get_final_damage(
            base_damage, 15, type_effectiveness, apply_burn, stab_mod, final_mod
        ),
    )


# NOTE: doesn't calculate pursuit or fling or naturalgift or battlebond
def calculate_base_power(
    attacker_identifier: str,
    defender_identifier: str,
    move: Move,
    battle: Union[Battle, DoubleBattle],
    has_ate_ability_type_change: bool,
):
    assert battle.player_role is not None and battle.opponent_role is not None
    attacker = battle.get_pokemon(attacker_identifier)
    defender = battle.get_pokemon(defender_identifier)

    attacker_spe = (attacker.stats["spe"] or 0) * (
        1.5 if attacker.item == "choicescarf" else 1
    )
    defender_spe = (defender.stats["spe"] or 0) * (
        1.5 if defender.item == "choicescarf" else 1
    )
    attacker_first = attacker_spe > defender_spe

    abilities_suppressed = any(
        map(lambda x: x and x.ability == "neutralizinggas", battle.all_active_pokemons)
    )

    base_power = move.base_power * 1.0

    defender_ability = defender.ability
    if move.id in MOVE_IGNORES_ABILITY or (
        attacker.ability in ATTACKER_IGNORES_ABILITY
        and not defender.item == "abilityshield"
    ):
        defender_ability = ""

    defender_weight = defender.weight
    if defender_ability == "lightmetal" and not abilities_suppressed:
        defender_weight *= 0.5
    elif defender_ability == "heavymetal" and not abilities_suppressed:
        defender_weight *= 2
    if defender.item == "floatstone":
        defender_weight *= 0.5

    attacker_weight = attacker.weight
    if attacker.ability == "lightmetal" and not abilities_suppressed:
        defender_weight *= 0.5
    elif attacker.ability == "heavymetal" and not abilities_suppressed:
        defender_weight *= 2
    if attacker.item == "floatstone":
        attacker_weight *= 0.5

    if move.id == "payback":
        base_power = base_power * (2 if not attacker_first else 1)
    elif move.id in ["boltbeak", "fishiousrend"]:
        base_power = base_power * (2 if attacker_first else 1)

    elif move.id == "electroball":
        if defender_spe == 0:
            base_power = 40
        elif math.floor(attacker_spe / defender_spe) >= 4:
            base_power = 150
        elif math.floor(attacker_spe / defender_spe) >= 3:
            base_power = 120
        elif math.floor(attacker_spe / defender_spe) >= 2:
            base_power = 80
        elif math.floor(attacker_spe / defender_spe) >= 1:
            base_power = 60
        else:
            base_power = 40
    elif move.id == "gyroball":
        if attacker_spe == 0:
            base_power = 1
        else:
            base_power = min(150, math.floor((25 * defender_spe) / attacker_spe) + 1)
    elif move.id == "punishment":
        base_power = min(
            200,
            60
            + 20
            * sum(
                map(lambda x: defender.boosts[x], ["atk", "def", "spa", "spd", "spe"])
            ),
        )
    elif move.id in ["lowkick", "grassknot"]:
        if defender_weight >= 200:
            base_power = 120
        elif defender_weight >= 100:
            base_power = 100
        elif defender_weight >= 50:
            base_power = 80
        elif defender_weight >= 25:
            base_power = 60
        elif defender_weight >= 10:
            base_power = 40
        else:
            base_power = 20
    elif move.id in ["hex", "infernalparade"]:
        base_power = base_power * (
            2 if defender.status is not None or defender.ability == "comatose" else 1
        )
    elif move.id == "barbarrage":
        base_power = base_power * (
            2 if defender.status in [Status.PSN, Status.TOX] else 1
        )
    elif move.id in ["heavyslam", "heatcrash"]:
        if defender_weight == 0:
            base_power = 40
        elif attacker_weight / defender_weight >= 5:
            base_power = 120
        elif attacker_weight / defender_weight >= 4:
            base_power = 100
        elif attacker_weight / defender_weight >= 3:
            base_power = 80
        elif attacker_weight / defender_weight >= 2:
            base_power = 60
        else:
            base_power = 40
    elif move.id in ["storedpower", "powertrip"]:
        base_power = 20 + 20 * sum(
            map(lambda x: attacker.boosts[x], ["atk", "def", "spa", "spd", "spe"])
        )
    elif move.id == "acrobatics":
        base_power = base_power * (
            2 if attacker.item == "flyinggem" or attacker.item is None else 1
        )
    elif move.id == "wakeupslap":
        base_power = base_power * (
            2 if defender.status == Status.SLP or defender.ability == "comatose" else 1
        )
    elif move.id == "smellsalts":
        base_power = base_power * (2 if defender.status == Status.PAR else 1)
    elif move.id == "weatherball":
        base_power = base_power * (2 if len(battle.weather) > 0 else 1)
        if attacker.item == "utilityumbrella" and (
            Weather.SUNNYDAY in battle.weather or Weather.RAINDANCE in battle.weather
        ):
            base_power = move.base_power
    elif move.id == "terrainpulse":
        base_power = base_power * (
            2
            if battle.is_grounded(attacker)
            and any(map(lambda x: x.is_terrain, battle.fields))
            else 1
        )
    elif move.id == "risingvoltage":
        base_power = base_power * (
            2
            if battle.is_grounded(defender) and Field.ELECTRIC_TERRAIN in battle.fields
            else 1
        )
    elif move.id == "psyblade":
        base_power = base_power * (
            1.5 if Field.ELECTRIC_TERRAIN in battle.fields else 1
        )
    elif move.id in ["dragonenergy", "eruption", "waterspout"]:
        base_power = max(1, math.floor(150 * attacker.current_hp_fraction))
    elif move.id in ["flail", "reversal"]:
        if math.floor(48 * attacker.current_hp_fraction) <= 1:
            base_power = 200
        elif math.floor(48 * attacker.current_hp_fraction) <= 4:
            base_power = 150
        elif math.floor(48 * attacker.current_hp_fraction) <= 9:
            base_power = 100
        elif math.floor(48 * attacker.current_hp_fraction) <= 16:
            base_power = 80
        elif math.floor(48 * attacker.current_hp_fraction) <= 32:
            base_power = 40
        else:
            base_power = 20
    elif move.id == "naturepower":
        if attacker.ability == "prankster" and PokemonType.DARK in defender.types:
            base_power = 0
        elif Field.ELECTRIC_TERRAIN in battle.fields:
            base_power = 90
        elif Field.GRASSY_TERRAIN in battle.fields:
            base_power = 90
        elif Field.MISTY_TERRAIN in battle.fields:
            base_power = 95
        elif Field.PSYCHIC_TERRAIN in battle.fields:
            if attacker.ability == "prankster" and battle.is_grounded(defender):
                base_power = 0
            else:
                base_power = 90
        else:
            base_power = 80
    elif move.id in ["crushgrip", "wringout", "hardpress"]:
        base_power = 100 * math.floor((defender.current_hp_fraction * 4096))
        base_power = math.floor(math.floor((120 * base_power + 2048 - 1) / 4096) / 100)
    elif move.id == "terablast":
        base_power = 100 if attacker.type_1 == PokemonType.STELLAR else 80

    if base_power == 0:
        return 0

    bp_mods = calculate_base_power_mods(
        attacker_identifier,
        defender_identifier,
        move,
        battle,
        base_power,
        has_ate_ability_type_change,
        attacker_first,
    )
    base_power = max(
        1, poke_round(base_power * chain_mods(bp_mods, 41, 2097152) / 4096)
    )

    # Tera moves have a minimum base power
    if (
        attacker.is_terastallized
        and move.type == attacker.tera_type
        and move.priority <= 0
        and move.base_power > 0
        and base_power < 60
    ) and move.id != "struggle":
        base_power = 60
    return base_power


def calculate_base_power_mods(
    attacker_identifier: str,
    defender_identifier: str,
    move: Move,
    battle: Union[Battle, DoubleBattle],
    base_power: float,
    has_ate_ability_type_change: bool,
    attacker_first: bool,  # replaces turn_order
):
    assert battle.player_role
    attacker = battle.get_pokemon(attacker_identifier)
    defender = battle.get_pokemon(defender_identifier)

    bp_mods = []
    abilities_suppressed = any(
        map(lambda x: x and x.ability == "neutralizinggas", battle.all_active_pokemons)
    )

    # ####################### Move Effects ########################
    # technically they need to be item/mon pairs, but this is good enough
    extra_knockoff_damage = defender.item is not None
    if defender.item is not None:
        if defender.item.endswith("memory"):
            extra_knockoff_damage = False
        elif defender.item.endswith("drive"):
            extra_knockoff_damage = False
        elif defender.item in {
            "adamantcrystal",
            "lustrousglobe",
            "griseousorb",
            "griseouscore",
            "redorb",
            "blueorb",
            "drive",
            "memory",
            "rustedsword",
            "rustedsheild",
            "cornerstonemask",
            "hearthflamemask",
            "wellspringmask",
            None,
        }:
            extra_knockoff_damage = False

    if (
        (
            move.id == "facade"
            and attacker.status in {Status.BRN, Status.PAR, Status.PSN, Status.TOX}
        )
        or (move.id == "brine" and defender.current_hp_fraction <= 0.5)
        or (move.id == "venoshock" and defender.status in {Status.PSN, Status.TOX})
        or (move.id == "lashout" and sum(attacker.boosts.values()) < 0)
    ):
        bp_mods.append(8192)
    elif (
        move.id == "expandingforce"
        and battle.is_grounded(attacker)
        and Field.PSYCHIC_TERRAIN in battle.fields
    ):
        bp_mods.append(6144)
    elif (
        (move.id == "knockoff" and extra_knockoff_damage)
        or (
            move.id == "mistyexplosion"
            and battle.is_grounded(attacker)
            and Field.MISTY_TERRAIN in battle.fields
        )
        or (move.id == "gravapple" and Field.GRAVITY in battle.fields)
    ):
        bp_mods.append(6144)
    elif (
        move.id in ["solarbeam", "solarblade"]
        and len(battle.weather) > 0
        and Weather.SUNNYDAY not in battle.fields
    ):
        bp_mods.append(2048)
    elif move.id in ["electrodrift", "collisioncourse"]:

        first_type_effectiveness = get_move_effectiveness(
            move,
            move.type,
            defender.type_1,
            attacker.ability in ["scrappy", "mindseye"],
            Field.GRAVITY in battle.fields,
            defender.item == "ringtarget",
        )

        second_type_effectiveness = (
            1
            if len(defender.types) == 1
            else get_move_effectiveness(
                move,
                move.type,
                defender.type_2,  # type: ignore
                attacker.ability in ["scrappy", "mindseye"],
                Field.GRAVITY in battle.fields,
                defender.item == "ringtarget",
            )
        )

        if first_type_effectiveness * second_type_effectiveness >= 2:
            bp_mods.append(5461)

    if Effect.HELPING_HAND in attacker.effects:
        bp_mods.append(6144)

    # ####################### Field Effects ########################

    if battle.is_grounded(attacker) and (
        (Field.ELECTRIC_TERRAIN in battle.fields and move.type == PokemonType.ELECTRIC)
        or (Field.GRASSY_TERRAIN in battle.fields and move.type == PokemonType.GRASS)
        or (Field.PSYCHIC_TERRAIN in battle.fields and move.type == PokemonType.PSYCHIC)
    ):
        bp_mods.append(5325)

    if battle.is_grounded(defender) and (
        (Field.MISTY_TERRAIN in battle.fields and move.type == PokemonType.DRAGON)
        or (
            Field.GRASSY_TERRAIN in battle.fields
            and move.id in ["bulldoze", "earthquake"]
        )
    ):
        bp_mods.append(2048)

    # ####################### Ability Effects ########################
    # NOTE: I don't support auras here

    if not abilities_suppressed and (
        (attacker.ability == "technician" and base_power <= 60)
        or (
            attacker.ability == "flareboost"
            and attacker.status == Status.BRN
            and move.category == MoveCategory.SPECIAL
        )
        or (
            attacker.ability == "toxicboost"
            and attacker.status in {Status.PSN, Status.TOX}
            and move.category == MoveCategory.PHYSICAL
        )
        or (attacker.ability == "megalauncher" and "pulse" in move.flags)
        or (attacker.ability == "strongjaw" and "bite" in move.flags)
        or (attacker.ability == "sharpness" and "slicing" in move.flags)
    ):
        bp_mods.append(6144)

    if not abilities_suppressed and (
        (attacker.ability == "sheerforce" and (move.secondary or move.id == "orderup"))
        or (
            attacker.ability == "sandforce"
            and Weather.SANDSTORM in battle.weather
            and move.type in {PokemonType.ROCK, PokemonType.GROUND, PokemonType.STEEL}
        )
        or (attacker.ability == "analytic" and not attacker_first)
        or (attacker.ability == "toughclaws" and "contact" in move.flags)
        or (attacker.ability == "punkrock" and "sound" in move.flags)
    ):
        bp_mods.append(5325)

    attacking_mons = (
        battle.active_pokemon
        if attacker_identifier.startswith(battle.player_role)
        else battle.opponent_active_pokemon
    )
    ally_attacking_mon = None
    if isinstance(battle, DoubleBattle) and isinstance(attacking_mons, list):
        ally_attacking_mon = battle.active_pokemon[
            1 if attacker == attacking_mons[0] else 0
        ]
    for mon in attacking_mons if isinstance(attacking_mons, list) else [attacking_mons]:
        if (
            mon is not None
            and mon.ability == "steelyspirit"
            and move.type == PokemonType.STEEL
            and not abilities_suppressed
        ):
            bp_mods.append(6144)

    if (
        ally_attacking_mon
        and ally_attacking_mon.ability == "battery"
        and move.category == MoveCategory.SPECIAL
        and not abilities_suppressed
    ):
        bp_mods.append(5325)

    if (
        ally_attacking_mon
        and ally_attacking_mon.ability == "powerspot"
        and not abilities_suppressed
    ):
        bp_mods.append(5325)

    if (
        attacker.ability == "rivalry"
        and not attacker.gender == PokemonGender.NEUTRAL
        and not defender.gender == PokemonGender.NEUTRAL
        and not abilities_suppressed
    ):
        if attacker.gender == defender.gender:
            bp_mods.append(5120)
        else:
            bp_mods.append(3072)

    if has_ate_ability_type_change:
        bp_mods.append(4915)

    if not abilities_suppressed and (
        (attacker.ability == "reckless" and move.recoil > 0)
        or (attacker.ability == "ironfist" and "punch" in move.flags)
    ):
        bp_mods.append(4915)

    if (
        defender.ability == "dryskin"
        and move.type == PokemonType.FIRE
        and not abilities_suppressed
    ):
        bp_mods.append(5120)

    attacking_team = (
        battle.team
        if attacker_identifier.startswith(battle.player_role)
        else battle.opponent_team
    )
    num_fainted = len([mon for mon in attacking_team.values() if mon.fainted])
    if (
        attacker.ability == "supremeoverlord"
        and num_fainted > 0
        and not abilities_suppressed
    ):
        pow_mod = [4096, 4506, 4915, 5325, 5734, 6144]
        bp_mods.append(pow_mod[min(num_fainted, 5)])

    # ####################### Item Effects ########################
    if (attacker.item or "").replace(" gem", "") == move.type.name.lower():
        bp_mods.append(5325)
    elif (
        (
            attacker.item == "adamantcrystal"
            and attacker.species == "dialgaorigin"
            and move.type in [PokemonType.STEEL, PokemonType.DRAGON]
        )
        or (
            attacker.item == "adamantorb"
            and attacker.species == "dialga"
            and move.type in [PokemonType.STEEL, PokemonType.DRAGON]
        )
        or (
            attacker.item == "lustrousorb"
            and attacker.species == "palkia"
            and move.type in [PokemonType.WATER, PokemonType.DRAGON]
        )
        or (
            attacker.item == "lustrousglobe"
            and attacker.species == "palkiaorigin"
            and move.type in [PokemonType.WATER, PokemonType.DRAGON]
        )
        or (
            attacker.item in ["griseousorb", "griseouscore"]
            and attacker.species in ["giratinaorigin", "giratina"]
            and move.type in [PokemonType.GHOST, PokemonType.DRAGON]
        )
        or (attacker.item == "souldew" and attacker.species in ["latios", "latias"])
        and move.type in [PokemonType.PSYCHIC, PokemonType.DRAGON]
        or (attacker.item and move.type == get_item_boost_type(attacker.item))
        or (
            attacker.species.startswith("ogerpon")
            and attacker.item
            and attacker.item.endswith("mask")
        )
    ):
        bp_mods.append(4915)
    elif (attacker.item == "muscleband" and move.category == MoveCategory.PHYSICAL) or (
        attacker.item == "wiseglasses" and move.category == MoveCategory.SPECIAL
    ):
        bp_mods.append(4505)
    elif attacker.item == "punchingglove" and "punch" in move.flags:
        bp_mods.append(4506)

    return bp_mods


def calculate_attack(
    attacker_identifier: str,
    defender_identifier: str,
    move: Move,
    battle: Union[Battle, DoubleBattle],
    is_critical: bool = False,
):
    attacker = battle.get_pokemon(attacker_identifier)
    defender = battle.get_pokemon(defender_identifier)

    attack = 0
    attack_stat = "atk" if move.category == MoveCategory.PHYSICAL else "spa"
    if move.id == "bodypress":
        attack_stat = "def"

    attack_source = attacker if move.id != "foulplay" else defender

    if (
        attack_source.stats[attack_stat] is not None
        and attack_source.boosts[attack_stat] == 0
        or (is_critical and attack_source.boosts[attack_stat] < 0)
        or defender.ability == "unaware"
    ):
        attack = attack_source.stats[attack_stat]  # type: ignore
    elif attack_source.stats[attack_stat] is not None:
        attack = math.floor(
            attack_source.stats[attack_stat]  # type: ignore
            * BOOST_MULTIPLIERS[attack_source.boosts[attack_stat]]
        )

    if attacker.ability == "hustle" and move.category == MoveCategory.PHYSICAL:
        attack = poke_round((attack * 3) / 2)  # type: ignore

    at_mods = calculate_atk_mods(attacker_identifier, defender_identifier, move, battle)
    return max(1, poke_round((attack * chain_mods(at_mods, 410, 131072)) / 4096))  # type: ignore


def calculate_atk_mods(
    attacker_identifier: str,
    defender_identifier: str,
    move: Move,
    battle: Union[Battle, DoubleBattle],
):
    assert battle.player_role is not None
    attacker = battle.get_pokemon(attacker_identifier)
    defender = battle.get_pokemon(defender_identifier)

    atk_mods = []
    attacker_mons = (
        battle.active_pokemon
        if attacker_identifier.startswith(battle.player_role)
        else battle.opponent_active_pokemon
    )
    attacker_ally = None
    if isinstance(battle, DoubleBattle) and isinstance(attacker_mons, list):
        attacker_ally = (
            attacker_mons[0] if attacker == attacker_mons[1] else attacker_mons[1]
        )

    abilities_suppressed = any(
        map(lambda x: x and x.ability == "neutralizinggas", battle.all_active_pokemons)
    )

    if (
        Effect.SLOW_START in attacker.effects and move.category == MoveCategory.PHYSICAL
    ) or (attacker.ability == "defeatist" and attacker.current_hp_fraction <= 0.5):
        atk_mods.append(2048)
    elif (
        attacker.ability == "solarpower"
        and Weather.SUNNYDAY in battle.weather
        and move.category == MoveCategory.SPECIAL
    ) or (
        attacker.species == "cherrim"
        and attacker.ability == "flowergift"
        and Weather.SUNNYDAY in battle.weather
        and move.category == MoveCategory.PHYSICAL
    ):
        atk_mods.append(6144)
    elif (
        attacker.ability == "gorillatactics" and move.category == MoveCategory.PHYSICAL
    ):
        atk_mods.append(6144)
    elif (
        (
            attacker.ability == "guts"
            and attacker.status
            and move.category == MoveCategory.PHYSICAL
        )
        or (
            attacker.current_hp_fraction <= 1.0 / 3
            and (
                (attacker.ability == "overgrow" and move.type == PokemonType.GRASS)
                or (attacker.ability == "blaze" and move.type == PokemonType.FIRE)
                or (attacker.ability == "torrent" and move.type == PokemonType.WATER)
                or (attacker.ability == "swarm" and move.type == PokemonType.BUG)
            )
        )
        or (
            isinstance(battle, DoubleBattle)
            and move.category == MoveCategory.SPECIAL
            and attacker_mons[0]  # type: ignore
            and attacker_mons[1]  # type: ignore
            and attacker_mons[0].ability in ["plus", "minus"]  # type: ignore
            and attacker_mons[1].ability in ["plus", "minus"]  # type: ignore
            and attacker_mons[0].ability != attacker_mons[1].ability  # type: ignore
            and not abilities_suppressed
        )
    ):
        atk_mods.append(6144)
    elif (
        attacker.ability == "flashfire"
        and Effect.FLASH_FIRE in attacker.effects
        and move.type == PokemonType.FIRE
    ):
        atk_mods.append(6144)
    elif (
        (attacker.ability == "steelworker" and move.type == PokemonType.STEEL)
        or (attacker.ability == "dragonsmaw" and move.type == PokemonType.DRAGON)
        or (attacker.ability == "rockypayload" and move.type == PokemonType.ROCK)
    ):
        atk_mods.append(6144)
    elif attacker.ability == "transistor" and move.type == PokemonType.ELECTRIC:
        atk_mods.append(5325)
    elif attacker.ability == "stakeout" and not abilities_suppressed:
        atk_mods.append(8192)
    elif (attacker.ability == "waterbubble" and move.type == PokemonType.WATER) or (
        attacker.ability in ["hugemount", "purepower"]
        and move.category == MoveCategory.PHYSICAL
    ):
        atk_mods.append(8192)

    if (
        attacker_ally
        and attacker_ally.ability == "flowergift"
        and Weather.SUNNYDAY in battle.weather
        and move.category == MoveCategory.PHYSICAL
        and not abilities_suppressed
    ):
        atk_mods.append(6144)

    if (
        attacker.ability == "steellyspirit"
        or (attacker_ally and attacker_ally.ability == "steellyspirit")
    ) and not abilities_suppressed:
        atk_mods.append(6144)

    if (
        (
            defender.ability == "thickfat"
            and move.type in [PokemonType.FIRE, PokemonType.ICE]
        )
        or (defender.ability == "waterbubble" and move.type == PokemonType.FIRE)
        or (defender.ability == "purifysalt" and move.type == PokemonType.GHOST)
    ):
        atk_mods.append(2048)

    if defender.ability == "heatproof" and move.type == PokemonType.FIRE:
        atk_mods.append(2048)

    is_tablets_of_ruin_active = (
        any(
            map(
                lambda x: x and x.ability == "tabletsofruin", battle.all_active_pokemons
            )
        )
        and not abilities_suppressed
    )

    is_vessel_of_ruin_active = (
        any(
            map(lambda x: x and x.ability == "vesselofruin", battle.all_active_pokemons)
        )
        and not abilities_suppressed
    )

    if is_tablets_of_ruin_active and move.category == MoveCategory.PHYSICAL:
        atk_mods.append(3072)

    if is_vessel_of_ruin_active and move.category == MoveCategory.SPECIAL:
        atk_mods.append(3072)

    if (
        (
            Effect.PROTOSYNTHESISATK in attacker.effects
            and move.category == MoveCategory.PHYSICAL
        )
        or (
            Effect.QUARKDRIVEATK in attacker.effects
            and move.category == MoveCategory.PHYSICAL
        )
        or (
            Effect.PROTOSYNTHESISSPA in attacker.effects
            and move.category == MoveCategory.SPECIAL
        )
        or (
            Effect.QUARKDRIVESPA in attacker.effects
            and move.category == MoveCategory.SPECIAL
        )
    ):
        atk_mods.append(5325)

    if (
        attacker.ability == "hadronengine"
        and move.category == MoveCategory.SPECIAL
        and Field.ELECTRIC_TERRAIN in battle.fields
    ) or (
        attacker.ability == "orichalcumpulse"
        and move.category == MoveCategory.PHYSICAL
        and Weather.SUNNYDAY in battle.weather
        and not attacker.item == "utilityumbrella"
    ):
        atk_mods.append(5461)

    if (
        (
            attacker.item == "thickclub"
            and attacker.base_species in ["cubone", "marowak"]
            and move.category == MoveCategory.PHYSICAL
        )
        or (
            attacker.item == "deepseatooth"
            and attacker.base_species == "clamperl"
            and move.category == MoveCategory.SPECIAL
        )
        or (attacker.item == "lightball" and attacker.base_species == "pikachu")
    ):
        atk_mods.append(8192)
    elif (attacker.item == "choiceband" and move.category == MoveCategory.PHYSICAL) or (
        attacker.item == "choicespecs" and move.category == MoveCategory.SPECIAL
    ):
        atk_mods.append(6144)

    return atk_mods


# NOTE: ignore shellsidearm
def calculate_defense(
    attacker_identifier: str,
    defender_identifier: str,
    move: Move,
    battle: Union[Battle, DoubleBattle],
    is_critical: bool = False,
):
    attacker = battle.get_pokemon(attacker_identifier)
    defender = battle.get_pokemon(defender_identifier)

    defense = 0

    hits_physical = (
        move.category == MoveCategory.PHYSICAL
        or move.entry.get("overrideDefensiveStat", "") == "def"
    )
    defense_stat = "def" if hits_physical else "spd"

    if (
        defender.stats[defense_stat] is not None
        and defender.boosts[defense_stat] == 0
        or (is_critical and defender.boosts[defense_stat] > 0)
        or move.ignore_defensive
        or attacker.ability == "unaware"
    ):
        defense = defender.stats[defense_stat]  # type: ignore
    elif defender.stats[defense_stat] is not None:
        defense = math.floor(
            defender.stats[defense_stat]  # type: ignore
            * BOOST_MULTIPLIERS[defender.boosts[defense_stat]]
        )

    if (
        Weather.SANDSTORM in battle.weather
        and PokemonType.ROCK in defender.types
        and not hits_physical
    ) or (
        Weather.SNOW in battle.weather
        and PokemonType.ICE in defender.types
        and hits_physical
    ):
        defense = poke_round((defense * 3) / 2)  # type: ignore

    def_mods = calculate_def_mods(
        attacker_identifier, defender_identifier, battle, hits_physical
    )

    return max(1, poke_round((defense * chain_mods(def_mods, 410, 131072)) / 4096))  # type: ignore


def calculate_def_mods(
    attacker_identifier: str,
    defender_identifier: str,
    battle: Union[Battle, DoubleBattle],
    hits_physical: bool = False,
):
    assert battle.player_role is not None
    attacker = battle.get_pokemon(attacker_identifier)
    defender = battle.get_pokemon(defender_identifier)

    defender_mons = (
        battle.active_pokemon
        if defender_identifier.startswith(battle.player_role)
        else battle.opponent_active_pokemon
    )
    defender_ally = None
    if isinstance(battle, DoubleBattle) and isinstance(defender_mons, list):
        defender_ally = (
            defender_mons[0] if defender == defender_mons[1] else defender_mons[1]
        )

    abilities_suppressed = any(
        map(lambda x: x and x.ability == "neutralizinggas", battle.all_active_pokemons)
    )

    def_mods = []
    if defender.ability == "marvelscale" and defender.status and hits_physical:
        def_mods.append(6144)
    elif (
        defender.species == "cherrim"
        and defender.ability == "flowergift"
        and Weather.SUNNYDAY not in battle.weather
        and not hits_physical
    ):
        def_mods.append(6144)
    elif (
        defender_ally
        and defender_ally.ability == "flowergift"
        and Weather.SUNNYDAY in battle.weather
        and not hits_physical
        and not abilities_suppressed
    ):
        def_mods.append(6144)
    elif (
        defender.ability == "grasspelt"
        and Field.GRASSY_TERRAIN in battle.fields
        and hits_physical
    ):
        def_mods.append(6144)
    elif defender.ability == "furcoat" and hits_physical:
        def_mods.append(8192)

    is_swords_of_ruin_active = (
        attacker.ability == "swordsofruin"
        or (
            any(
                map(
                    lambda x: x and x.ability == "swordsofruin",
                    battle.all_active_pokemons,
                )
            )
            and defender.ability != "swordsofruin"
        )
    ) and not abilities_suppressed

    is_beads_of_ruin_active = (
        attacker.ability == "beadsofruin"
        or (
            any(
                map(
                    lambda x: x and x.ability == "beadsofruin",
                    battle.all_active_pokemons,
                )
            )
            and defender.ability != "beadsofruin"
        )
    ) and not abilities_suppressed

    if (is_swords_of_ruin_active and hits_physical) or (
        is_beads_of_ruin_active and not hits_physical
    ):
        def_mods.append(3072)

    if (
        hits_physical
        and (
            Effect.QUARKDRIVEDEF in defender.effects
            or Effect.PROTOSYNTHESISDEF in defender.effects
        )
    ) or (
        not hits_physical
        and (
            Effect.QUARKDRIVESPD in defender.effects
            or Effect.PROTOSYNTHESISSPD in defender.effects
        )
    ):
        def_mods.append(5324)

    if (
        defender.item == "eviolite"
        and (
            defender.species == "dipplin"
            or len(
                GenData.from_gen(battle.gen).pokedex[defender.species].get("evos", [])
            )
            > 0
        )
    ) or (defender.item == "assaultvest" and not hits_physical):
        def_mods.append(6144)
    elif (
        defender.item == "metalpowder" and defender.species == "ditto" and hits_physical
    ) or (
        defender.item == "deepseascale"
        and defender.species == "clamperl"
        and not hits_physical
    ):
        def_mods.append(8192)

    return def_mods


def calculate_base_damage(
    attacker_identifier: str,
    defender_identifier: str,
    base_power: float,
    attack: float,
    defense: float,
    move: Move,
    move_type: Optional[PokemonType],
    move_target: Target,
    battle: Union[Battle, DoubleBattle],
    is_critical: bool = False,
):
    assert battle.player_role is not None
    attacker = battle.get_pokemon(attacker_identifier)
    defender = battle.get_pokemon(defender_identifier)
    defending_mons = (
        battle.active_pokemon
        if defender_identifier.startswith(battle.player_role)
        else battle.opponent_active_pokemon
    )

    is_spread = (
        isinstance(battle, DoubleBattle)
        and defending_mons[0] is not None  # type: ignore
        and defending_mons[1] is not None  # type: ignore
        and move_target in [Target.ALL_ADJACENT, Target.ALL_ADJACENT_FOES]
    )

    base_damage = math.floor(
        math.floor(
            (math.floor(2 * attacker.level / 5 + 2) * base_power) * attack / defense
        )
        / 50
        + 2
    )

    if is_spread:
        base_damage = poke_round((base_damage * 3072) / 4096)

    if (
        Weather.SUNNYDAY in battle.weather
        and move.id == "hydrosteam"
        and attacker.item != "utilityumbrella"
    ):
        base_damage = poke_round((base_damage * 6144) / 4096)
    elif defender.item != "utilityumbrella":
        if (Weather.SUNNYDAY in battle.weather and move_type == PokemonType.FIRE) or (
            Weather.RAINDANCE in battle.weather and move_type == PokemonType.WATER
        ):
            base_damage = poke_round((base_damage * 6144) / 4096)
        elif (
            Weather.SUNNYDAY in battle.weather and move_type == PokemonType.WATER
        ) or (Weather.RAINDANCE in battle.weather and move_type == PokemonType.FIRE):
            base_damage = poke_round((base_damage * 2048) / 4096)

    if is_critical:
        base_damage = poke_round((base_damage * 1.5))

    return base_damage


# NOTE: do not calculate metronome
def calculate_final_mods(
    attacker_identifier: str,
    defender_identifier: str,
    move: Move,
    battle: Union[Battle, DoubleBattle],
    type_effectiveness: float,
    flags: Dict[str, int],
    is_critical: bool = False,
    hit_count: int = 0,
):
    assert battle.player_role is not None
    attacker = battle.get_pokemon(attacker_identifier)
    defender = battle.get_pokemon(defender_identifier)
    final_mods = []

    defender_side_conditions = (
        battle.side_conditions
        if defender_identifier.startswith(battle.player_role)
        else battle.opponent_side_conditions
    )
    defending_mons = (
        battle.active_pokemon
        if defender_identifier.startswith(battle.player_role)
        else battle.opponent_active_pokemon
    )
    defender_ally = None
    if isinstance(defending_mons, list):
        defender_ally = (
            defending_mons[0] if defender == defending_mons[1] else defending_mons[1]
        )

    abilities_suppressed = any(
        map(lambda x: x and x.ability == "neutralizinggas", battle.all_active_pokemons)
    )

    if (
        SideCondition.REFLECT in defender_side_conditions
        and move.category == MoveCategory.PHYSICAL
        and not is_critical
        and SideCondition.AURORA_VEIL not in defender_side_conditions
        and not flags.get("ignorescreens", 0) == 1
    ):
        final_mods.append(2048 if not isinstance(battle, DoubleBattle) else 2732)
    elif (
        SideCondition.LIGHT_SCREEN in defender_side_conditions
        and move.category == MoveCategory.SPECIAL
        and not is_critical
        and SideCondition.AURORA_VEIL not in defender_side_conditions
        and not flags.get("ignorescreens", 0) == 1
    ):
        final_mods.append(2048 if not isinstance(battle, DoubleBattle) else 2732)
    elif (
        SideCondition.AURORA_VEIL in defender_side_conditions
        and not is_critical
        and not flags.get("ignorescreens", 0) == 1
    ):
        final_mods.append(2048 if not isinstance(battle, DoubleBattle) else 2732)

    if attacker.ability == "neuroforce" and type_effectiveness > 1:
        final_mods.append(5120)
    elif attacker.ability == "sniper" and is_critical:
        final_mods.append(6144)
    elif attacker.ability == "tintedlens" and type_effectiveness < 1:
        final_mods.append(8192)

    if (
        defender.ability in ["multiscale", "shadowshield"]
        and defender.current_hp_fraction == 1
        and hit_count <= 1
    ):
        final_mods.append(2048)

    if (
        defender.ability == "fluffy"
        and "contact" in flags
        and not attacker.ability == "longreach"
    ):
        final_mods.append(2048)
    elif (defender.ability == "punkrock" and "sound" in flags) or (
        defender.ability == "icescales" and move.category == MoveCategory.SPECIAL
    ):
        final_mods.append(2048)

    if (
        defender.ability in ["solidrock", "filter", "prismarmor"]
        and type_effectiveness > 1
    ):
        final_mods.append(3072)

    if (
        defender_ally
        and defender_ally.ability == "friendguard"
        and not abilities_suppressed
    ):
        final_mods.append(3072)

    if defender.ability == "fluffy" and move.type == PokemonType.FIRE:
        final_mods.append(8192)

    if defender.ability == "fluffy" and move.type == PokemonType.FIRE:
        final_mods.append(8192)

    if attacker.item == "expertbelt" and type_effectiveness > 1:
        final_mods.append(4915)
    elif attacker.item == "lifeorb":
        final_mods.append(5324)

    if (
        (defender.item and get_berry_resist_type(defender.item) == move.type)
        and (type_effectiveness > 1 or PokemonType.NORMAL)
        and hit_count == 0
        and attacker.ability not in ["unnerve", "asone"]
    ):
        final_mods.append(1024 if defender.ability == "ripen" else 2048)

    return final_mods


def get_stab_mod(pokemon: Pokemon, move: Move, move_type: Optional[PokemonType]):
    stab_mod = 4096
    if move.id == "struggle":  # struggle is typeless
        return stab_mod

    if move_type in pokemon.base_types:
        stab_mod += 2048
    elif pokemon.ability in ["protean", "libero"] and not pokemon.is_terastallized:
        stab_mod += 2048

    tera_type = pokemon.tera_type
    if (
        tera_type == move_type
        and tera_type != PokemonType.STELLAR
        and pokemon.is_terastallized
    ):
        stab_mod += 2048
    if pokemon.ability == "adaptability" and move_type in pokemon.types:
        stab_mod += (
            1024
            if pokemon.is_terastallized and tera_type in pokemon.base_types
            else 2048
        )

    return stab_mod


# NOTE: we don't store whether a pokemon used a first move after stellar, and so
# we fail all stellar calculations
def get_stellar_stab_mod(pokemon: Pokemon, move: Move, stab_mod=1):
    is_stellar_boosted = False
    # is_stellar_boosted = (
    #     pokemon.tera_type == PokemonType.STELLAR and (
    #         (move.is_stellar_first_use and turns == 0) or pokemon.species == "terapagosstellar"
    #     )
    # )

    if is_stellar_boosted:
        if move.type in pokemon.base_types:
            stab_mod += 2048
        else:
            stab_mod = 4915

    return stab_mod


def get_berry_resist_type(berry: str) -> Optional[PokemonType]:
    return BERRY_RESISTS.get(berry, None)


def get_item_boost_type(item: str) -> Optional[PokemonType]:
    return ITEM_BOOST_TYPES.get(item, None)  # type: ignore


# Gamefreak rounds down on even number
def poke_round(num: float):
    return math.ceil(num) if num % 1 > 0.5 else math.floor(num)


def get_final_damage(
    base_amount: int,
    i: int,
    effectiveness: float,
    is_burned: bool,
    stab_mod: int,
    final_mod: int,
) -> float:
    damage_amount = math.floor(base_amount * (85 + i) / 100) * 1.0
    if stab_mod != 4096:
        damage_amount = (damage_amount * stab_mod) / 4096

    damage_amount = math.floor(poke_round(damage_amount) * effectiveness)

    if is_burned:
        damage_amount = math.floor(damage_amount / 2)

    return poke_round(max(1, (damage_amount * final_mod) / 4096))


def chain_mods(mods: List[int], lb: int, ub: int):
    m = 4096
    for mod in mods:
        if mod != 4096:
            m = (m * mod + 2048) >> 12

    return max(min(m, ub), lb)


def get_move_effectiveness(
    move: Move,
    move_type: PokemonType,
    type: PokemonType,
    is_ghost_revealed: Optional[bool] = False,
    is_gravity: Optional[bool] = False,
    is_ring_target: Optional[bool] = False,
) -> float:
    if (
        is_ghost_revealed
        and type == PokemonType.GHOST
        and move_type in (PokemonType.NORMAL, PokemonType.FIGHTING)
    ):
        return 1
    elif move.id == "struggle":
        return 1
    elif is_gravity and type == PokemonType.FLYING and move_type == PokemonType.GROUND:
        return 1
    elif move.id == "freezedry" and type == PokemonType.WATER:
        return 2
    else:
        effectiveness = PokemonType.damage_multiplier(
            move_type, type, type_chart=GenData.from_gen(9).type_chart
        )
        if effectiveness == 0 and is_ring_target:
            effectiveness = 1
        if move.id == "flyingpress":
            effectiveness *= PokemonType.damage_multiplier(
                PokemonType.FLYING, type, type_chart=GenData.from_gen(9).type_chart
            )
        return effectiveness


PROTECT_EFFECTS = {
    Effect.PROTECT,
    Effect.SPIKY_SHIELD,
    Effect.KINGS_SHIELD,
    Effect.BANEFUL_BUNKER,
    Effect.BURNING_BULWARK,
    Effect.OBSTRUCT,
    Effect.MAX_GUARD,
    Effect.SILK_TRAP,
}

ABILITY_IGNORES_NEUTRALIZING_GAS = {
    "asone",
    "battlebond",
    "comatose",
    "disguise",
    "gulpmissile",
    "iceface",
    "multitype",
    "neutralizinggas",
    "powerconstruct",
    "rksystem",
    "schooling",
    "shieldsdown",
    "stancechange",
    "terashift",
    "zenmode",
    "zerotohero",
}

ATTACKER_IGNORES_ABILITY = {"moldbreaker", "teravolt", "turboblaze"}

MOVE_IGNORES_ABILITY = {"moongeistbeam", "photongeyser", "sunsteelstrike"}

DEFENDER_ABILITY_IGNORED = {
    "armortail",
    "aromaveil",
    "aurabreak",
    "battlearmor",
    "bigpecks",
    "bulletproof",
    "clearbody",
    "contrary",
    "damp",
    "dazzling",
    "disguise",
    "dryskin",
    "eartheater",
    "filter",
    "flashfire",
    "flowergift",
    "flowerveil",
    "fluffy",
    "friendguard",
    "furcoat",
    "goodasgold",
    "grasspelt",
    "guarddog",
    "heatproof",
    "heavymetal",
    "hypercutter",
    "iceface",
    "icescales",
    "illuminate",
    "immunity",
    "innerfocus",
    "insomnia",
    "keeneye",
    "leafguard",
    "levitate",
    "lightmetal",
    "lightningrod",
    "limber",
    "magicbounce",
    "magmarmor",
    "marvelscale",
    "mindseye",
    "mirrorarmor",
    "motordrive",
    "multiscale",
    "oblivious",
    "overcoat",
    "owntempo",
    "pastelveil",
    "punkrock",
    "purifysalt",
    "queenlymajesty",
    "quickfeet",
    "raindish",
    "rockhead",
    "rocky",
    "roughskin",
    "sandveil",
    "scrappy",
    "simple",
    "snowcloak",
    "snowwarning",
    "solidrock",
    "soundproof",
    "speedboost",
    "stakeout",
    "steelworker",
    "strongjaw",
    "suctioncups",
    "tintedlens",
    "torrent",
    "trace",
    "waterabsorb",
    "waterveil",
    "weakarmor",
    "wonderguard",
    "wonderskin",
}

BOOST_MULTIPLIERS = {
    -6: 2.0 / 8,
    -5: 2.0 / 7,
    -4: 2.0 / 6,
    -3: 2.0 / 5,
    -2: 2.0 / 4,
    -1: 2.0 / 3,
    0: 1,
    1: 1.5,
    2: 2,
    3: 2.5,
    4: 3,
    5: 3.5,
    6: 4,
}

BERRY_RESISTS = {
    "chilanberry": PokemonType.NORMAL,
    "occaberry": PokemonType.FIRE,
    "passhoberry": PokemonType.WATER,
    "wacanberry": PokemonType.ELECTRIC,
    "rindoberry": PokemonType.GRASS,
    "yacheberry": PokemonType.ICE,
    "chopleberry": PokemonType.FIGHTING,
    "kebiaberry": PokemonType.POISON,
    "shucaberry": PokemonType.GROUND,
    "cobaberry": PokemonType.FLYING,
    "payapaberry": PokemonType.PSYCHIC,
    "tangaberry": PokemonType.BUG,
    "chartiberry": PokemonType.ROCK,
    "kasiberry": PokemonType.GHOST,
    "habanberry": PokemonType.DRAGON,
    "colburberry": PokemonType.DARK,
    "babiriberry": PokemonType.STEEL,
    "roseliberry": PokemonType.FAIRY,
}

ITEM_BOOST_TYPES = {
    "dracoplate": PokemonType.DRAGON,
    "dragonfang": PokemonType.DRAGON,
    "dreadplate": PokemonType.DARK,
    "blackglasses": PokemonType.DARK,
    "earthplate": PokemonType.GROUND,
    "softsand": PokemonType.GROUND,
    "fistplate": PokemonType.FIGHTING,
    "blackbelt": PokemonType.FIGHTING,
    "flameplate": PokemonType.FIRE,
    "charcoal": PokemonType.FIRE,
    "icicleplate": PokemonType.ICE,
    "nevermeltice": PokemonType.ICE,
    "insectplate": PokemonType.BUG,
    "silverpowder": PokemonType.BUG,
    "ironplate": PokemonType.STEEL,
    "metalcoat": PokemonType.STEEL,
    "meadowplate": PokemonType.GRASS,
    "roseincense": PokemonType.GRASS,
    "miracleseed": PokemonType.GRASS,
    "mindplate": PokemonType.PSYCHIC,
    "oddincense": PokemonType.PSYCHIC,
    "twistedspoon": PokemonType.PSYCHIC,
    "fairyfeather": PokemonType.FAIRY,
    "pixieplate": PokemonType.FAIRY,
    "skyplate": PokemonType.FLYING,
    "sharpbeak": PokemonType.FLYING,
    "splashplate": PokemonType.WATER,
    "seaincense": PokemonType.WATER,
    "waveincense": PokemonType.WATER,
    "mysticwater": PokemonType.WATER,
    "spookyplate": PokemonType.GHOST,
    "spelltag": PokemonType.GHOST,
    "stoneplate": PokemonType.ROCK,
    "rockincense": PokemonType.ROCK,
    "hardstone": PokemonType.ROCK,
    "toxicplate": PokemonType.POISON,
    "poisonbarb": PokemonType.POISON,
    "zapplate": PokemonType.ELECTRIC,
    "magnet": PokemonType.ELECTRIC,
    "silkscarf": PokemonType.NORMAL,
    "pinkbow": PokemonType.NORMAL,
    "polkadotbow": PokemonType.NORMAL,
}



from typing import Any, Optional, Union

import numpy as np
from gymnasium.spaces import Discrete, Space

from poke_env.battle import Battle, Pokemon
from poke_env.battle.move import SPECIAL_MOVES
from poke_env.data import GenData
from poke_env.environment.env import PokeEnv
from poke_env.player.battle_order import (
    BattleOrder,
    DefaultBattleOrder,
    ForfeitBattleOrder,
    SingleBattleOrder,
)
from poke_env.player.player import Player
from poke_env.ps_client import (
    AccountConfiguration,
    LocalhostServerConfiguration,
    ServerConfiguration,
)
from poke_env.teambuilder import Teambuilder


class SinglesEnv(PokeEnv[np.int64]):
    def __init__(
        self,
        *,
        account_configuration1: Optional[AccountConfiguration] = None,
        account_configuration2: Optional[AccountConfiguration] = None,
        avatar: Optional[int] = None,
        battle_format: str = "gen8randombattle",
        log_level: Optional[int] = None,
        save_replays: Union[bool, str] = False,
        server_configuration: Optional[
            ServerConfiguration
        ] = LocalhostServerConfiguration,
        accept_open_team_sheet: Optional[bool] = False,
        start_timer_on_battle_start: bool = False,
        start_listening: bool = True,
        open_timeout: Optional[float] = 10.0,
        ping_interval: Optional[float] = 20.0,
        ping_timeout: Optional[float] = 20.0,
        challenge_timeout: Optional[float] = 60.0,
        team: Optional[Union[str, Teambuilder]] = None,
        choose_on_teampreview: bool = True,
        fake: bool = False,
        strict: bool = True,
    ):
        super().__init__(
            account_configuration1=account_configuration1,
            account_configuration2=account_configuration2,
            avatar=avatar,
            battle_format=battle_format,
            log_level=log_level,
            save_replays=save_replays,
            server_configuration=server_configuration,
            accept_open_team_sheet=accept_open_team_sheet,
            start_timer_on_battle_start=start_timer_on_battle_start,
            start_listening=start_listening,
            open_timeout=open_timeout,
            ping_interval=ping_interval,
            ping_timeout=ping_timeout,
            challenge_timeout=challenge_timeout,
            team=team,
            choose_on_teampreview=choose_on_teampreview,
            fake=fake,
            strict=strict,
        )
        gen = GenData.from_format(battle_format).gen
        self.action_spaces: dict[str, Space[Any]] = {
            agent: Discrete(SinglesEnv.get_action_space_size(gen))
            for agent in self.possible_agents
        }

    @staticmethod
    def action_to_order(
        action: np.int64, battle: Battle, fake: bool = False, strict: bool = True
    ) -> BattleOrder:
        """
        Returns the BattleOrder relative to the given action.

        The action mapping is as follows:
        action = -2: default
        action = -1: forfeit
        0 <= action <= 5: switch
        6 <= action <= 9: move
        10 <= action <= 13: move and mega evolve
        14 <= action <= 17: move and z-move
        18 <= action <= 21: move and dynamax
        22 <= action <= 25: move and terastallize

        :param action: The action to take.
        :type action: int64
        :param battle: The current battle state
        :type battle: AbstractBattle
        :param fake: If true, action-order converters will try to avoid returning a default
            output if at all possible, even if the output isn't a legal decision. Defaults
            to False.
        :type fake: bool
        :param strict: If true, action-order converters will throw an error if the move is
            illegal. Otherwise, it will return default. Defaults to True.
        :type strict: bool

        :return: The battle order for the given action in context of the current battle.
        :rtype: BattleOrder
        """
        try:
            if action == -2:
                return DefaultBattleOrder()
            elif action == -1:
                return ForfeitBattleOrder()
            elif action < 6:
                order = Player.create_order(list(battle.team.values())[action])
            else:
                if battle.active_pokemon is None:
                    raise ValueError(
                        f"Invalid order from player {battle.player_username} "
                        f"in battle {battle.battle_tag} - action specifies a "
                        f"move, but battle.active_pokemon is None!"
                    )
                avail_ids = [m.id for m in battle.available_moves]
                known_moves = list(battle.active_pokemon.moves.values())[:4]
                known_ids = [m.id for m in known_moves]
                mvs = (
                    battle.available_moves
                    if len(avail_ids) == 1 and avail_ids[0] not in known_ids
                    else known_moves
                )
                if (action - 6) % 4 not in range(len(mvs)):
                    raise ValueError(
                        f"Invalid action {action} from player {battle.player_username} "
                        f"in battle {battle.battle_tag} - action specifies a move "
                        f"but the move index {(action - 6) % 4} is out of bounds "
                        f"for available moves {mvs}!"
                    )
                order = Player.create_order(
                    mvs[(action - 6) % 4],
                    mega=10 <= action.item() < 14,
                    z_move=14 <= action.item() < 18,
                    dynamax=18 <= action.item() < 22,
                    terastallize=22 <= action.item() < 26,
                )
            if not fake and str(order) not in [str(o) for o in battle.valid_orders]:
                raise ValueError(
                    f"Invalid action {action} from player {battle.player_username} "
                    f"in battle {battle.battle_tag} - converted order {order} "
                    f"not in valid orders {[str(o) for o in battle.valid_orders]}!"
                )
            return order
        except ValueError as e:
            if strict:
                raise e
            else:
                if battle.logger is not None:
                    battle.logger.warning(str(e) + " Defaulting to random move.")
                return Player.choose_random_singles_move(battle)

    @staticmethod
    def order_to_action(
        order: BattleOrder, battle: Battle, fake: bool = False, strict: bool = True
    ) -> np.int64:
        """
        Returns the action relative to the given BattleOrder.

        :param order: The order to take.
        :type order: BattleOrder
        :param battle: The current battle state
        :type battle: AbstractBattle
        :param fake: If true, action-order converters will try to avoid returning a default
            output if at all possible, even if the output isn't a legal decision. Defaults
            to False.
        :type fake: bool
        :param strict: If true, action-order converters will throw an error if the move is
            illegal. Otherwise, it will return default. Defaults to True.
        :type strict: bool

        :return: The action for the given battle order in context of the current battle.
        :rtype: int64
        """
        try:
            if isinstance(order, DefaultBattleOrder):
                action = -2
            elif isinstance(order, ForfeitBattleOrder):
                action = -1
            else:
                assert isinstance(order, SingleBattleOrder)
                assert not isinstance(order.order, str)
                if not fake and str(order) not in [str(o) for o in battle.valid_orders]:
                    raise ValueError(
                        f"Invalid order from player {battle.player_username} "
                        f"in battle {battle.battle_tag} - order {order} "
                        f"not in valid orders {[str(o) for o in battle.valid_orders]}!"
                    )
                if isinstance(order.order, Pokemon):
                    action = [p.base_species for p in battle.team.values()].index(
                        order.order.base_species
                    )
                else:
                    assert battle.active_pokemon is not None
                    avail_ids = [m.id for m in battle.available_moves]
                    known_moves = list(battle.active_pokemon.moves.values())[:4]
                    known_ids = [m.id for m in known_moves]
                    mvs = (
                        battle.available_moves
                        if len(avail_ids) == 1 and avail_ids[0] not in known_ids
                        else known_moves
                    )
                    action = [m.id for m in mvs].index(order.order.id)
                    if order.mega:
                        gimmick = 1
                    elif order.z_move:
                        gimmick = 2
                    elif order.dynamax:
                        gimmick = 3
                    elif order.terastallize:
                        gimmick = 4
                    else:
                        gimmick = 0
                    action = 6 + action + 4 * gimmick
            return np.int64(action)
        except ValueError as e:
            if strict:
                raise e
            else:
                if battle.logger is not None:
                    battle.logger.warning(str(e) + " Defaulting to random move.")
                return SinglesEnv.order_to_action(
                    Player.choose_random_singles_move(battle), battle, fake, strict
                )

    @staticmethod
    def get_action_mask(battle: Battle) -> list[int]:
        switch_space = [
            i
            for i, pokemon in enumerate(battle.team.values())
            if not battle.trapped
            and pokemon.base_species
            in [p.base_species for p in battle.available_switches]
        ]
        if battle._wait:
            actions = [0]
        elif battle.active_pokemon is None:
            actions = switch_space
        else:
            known_moves = list(battle.active_pokemon.moves.values())[:4]
            move_space = [
                i + 6
                for i, move in enumerate(known_moves)
                if move.id in [m.id for m in battle.available_moves]
            ]
            if (
                not move_space
                and len(battle.available_moves) == 1
                and battle.available_moves[0].id not in SPECIAL_MOVES
            ):
                move_space = [6]
            mega_space = [i + 4 for i in move_space if battle.can_mega_evolve]
            zmove_space = [
                i + 8
                for i in move_space
                if known_moves[i - 6].id
                in [m.id for m in battle.active_pokemon.available_z_moves]
                and battle.can_z_move
            ]
            dynamax_space = [i + 12 for i in move_space if battle.can_dynamax]
            tera_space = [i + 16 for i in move_space if battle.can_tera]
            if (
                not move_space
                and len(battle.available_moves) == 1
                and battle.available_moves[0].id in SPECIAL_MOVES
            ):
                move_space = [6]
            actions = (
                switch_space
                + move_space
                + mega_space
                + zmove_space
                + dynamax_space
                + tera_space
            )
        action_mask = [
            int(i in actions)
            for i in range(SinglesEnv.get_action_space_size(battle.gen))
        ]
        return action_mask

    @staticmethod
    def get_action_space_size(gen: int) -> int:
        num_switches = 6
        num_moves = 4
        if gen == 6:
            num_gimmicks = 1
        elif gen == 7:
            num_gimmicks = 2
        elif gen == 8:
            num_gimmicks = 3
        elif gen == 9:
            num_gimmicks = 4
        else:
            num_gimmicks = 0
        return num_switches + num_moves * (num_gimmicks + 1)



from typing import Any, Awaitable, Dict, Optional, Tuple

import numpy as np
import numpy.typing as npt
from gymnasium import Env
from gymnasium.envs.registration import EnvSpec

from poke_env.environment.env import ActionType, PokeEnv
from poke_env.player.player import Player


class SingleAgentWrapper(Env[Dict[str, Any], ActionType]):
    def __init__(self, env: PokeEnv[ActionType], opponent: Player):
        self.env = env
        self.opponent = opponent
        self.spec = EnvSpec("poke-env-v0", nondeterministic=True)
        self.observation_space = env.observation_spaces[env.agent1.username]
        self.action_space = env.action_spaces[env.agent1.username]
        self.second_teampreview_action: npt.NDArray[np.int64] | None = None

    def step(
        self, action: ActionType
    ) -> Tuple[Dict[str, Any], float, bool, bool, Dict[str, Any]]:
        """Run one timestep of the environment's dynamics.

        Takes an action from the main agent and automatically generates an action
        for the opponent using the opponent's choose_move method.

        :param action: Action from the main agent.
        :type action: ActionType
        :return: Tuple of (observation, reward, terminated, truncated, info) for the main agent.
        :rtype: Tuple[Dict[str, Any], float, bool, bool, Dict[str, Any]]
        """
        assert self.env.battle2 is not None
        if not self.env.battle2.teampreview:
            opp_order = self.opponent.choose_move(self.env.battle2)
            assert not isinstance(opp_order, Awaitable)
            opp_action = self.env.order_to_action(
                opp_order,
                self.env.battle2,
                fake=self.env._fake,
                strict=self.env._strict,
            )
        elif self.env.battle2.format is None or "vgc" not in self.env.battle2.format:
            raise NotImplementedError(
                "Teampreview is only supported for VGC formats in SingleAgentWrapper."
            )
        elif self.second_teampreview_action is None:
            tp_order = self.opponent.teampreview(self.env.battle2)
            assert not isinstance(tp_order, Awaitable)
            assert len(tp_order) == 10, f"{tp_order} must specify 4 slots in VGC!"
            teampreview_order_list = [int(i) for i in tp_order[-4:]]
            opp_action = np.array(teampreview_order_list[:2])  # type: ignore
            self.second_teampreview_action = np.array(teampreview_order_list[2:])
            # only the first two pokemon are selected in teampreview for now
            for i, pokemon in enumerate(self.env.battle2.team.values(), start=1):
                pokemon._selected_in_teampreview = i in teampreview_order_list[:2]
        else:
            opp_action = self.second_teampreview_action  # type: ignore
            # now the second two pokemon are selected in teampreview
            for i in opp_action:  # type: ignore
                mon = list(self.env.battle2.team.values())[i - 1]
                mon._selected_in_teampreview = True
            self.second_teampreview_action = None
        actions = {
            self.env.agent1.username: action,
            self.env.agent2.username: opp_action,
        }
        obs, rewards, terms, truncs, infos = self.env.step(actions)
        return (
            obs[self.env.agent1.username],
            rewards[self.env.agent1.username],
            terms[self.env.agent1.username],
            truncs[self.env.agent1.username],
            infos[self.env.agent1.username],
        )

    def reset(
        self, *, seed: Optional[int] = None, options: Optional[Dict[str, Any]] = None
    ) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        obs, infos = self.env.reset(seed, options)
        self.opponent.reset_battles()
        assert self.env.battle2 is not None
        self.opponent._battles[self.env.battle2.battle_tag] = self.env.battle2
        self._np_random = self.env._np_random
        return obs[self.env.agent1.username], infos[self.env.agent1.username]

    def render(self, mode="human"):
        return self.env.render(mode)

    def close(self):
        self.env.close()
